/* HSM Object Management Functions */

#include <strings.h>
#include <libpki/pki.h>


/* \brief Allocates a new HSM structure
 *
 * Allocates a new HSM structure and initialize the callbacks functions.
 * The driver is the crypto driver to be used (e.g., openssl or kmf),
 * while the name is the name of the HSM (e.g., LunaCA3)
 */

HSM *HSM_new( char *dir, char *name ) {

	HSM *hsm = NULL;
	char * url_s = NULL;
	char * buff = NULL;

	PKI_CONFIG *conf = NULL;
	char *type = NULL;

	if( !dir || !name ) return (NULL);

	if((url_s = PKI_CONFIG_find( dir, name )) == NULL ) {
		return (NULL);
	}

	if((conf = PKI_CONFIG_load( url_s )) == NULL ) {
		return(NULL);
	}

	if((buff = PKI_Malloc ( BUFF_MAX_SIZE )) == NULL ) {
		return(NULL);
	}

	/* Let's generate the right searching string with the namespace
	   prefix */
	if((type = PKI_CONFIG_get_value ( conf, "/hsm/type")) == NULL ) {
		/* No type in the config! */
		PKI_log_debug("ERROR, No HSM type in the config!");
		PKI_CONFIG_free ( conf );
		return (NULL);
	}

	if( strcmp_nocase(type,"software") == 0 ) {
		if((hsm = HSM_OPENSSL_new( conf )) == NULL ) {
			PKI_log_debug("ERROR, Can not generate software HSM object!");
		} else {
			hsm->type = HSM_TYPE_SOFTWARE;
		}
	} else if( strcmp_nocase(type,"engine") == 0 ) {
		if((hsm = HSM_ENGINE_new( conf )) == NULL ) {
			PKI_log_debug("ERROR, Can not generate engine HSM object!");
		} else {
			hsm->type = HSM_TYPE_ENGINE;
		}
	} else if( strcmp_nocase(type,"kmf") == 0 ) {
#ifdef ENABLE_KMF
		if((hsm = HSM_KMF_new( conf )) == NULL ) {
			PKI_log_debug("ERROR, Can not generate kmf HSM object!\n");
		} else {
			hsm->type = HSM_TYPE_KMF;
		}
#endif
	} else {
		PKI_CONFIG_free ( conf );
		return (NULL);
	}

	return (hsm);
}

int HSM_free ( HSM *hsm ) {

	if( !hsm ) return (PKI_ERR);

	if( hsm && hsm->callbacks->free ) {
		hsm->callbacks->free ( (void *) hsm->driver );
	} else {
		/* Error! The driver should provide a free callback! */
		PKI_log_debug("ERROR, hsm (%s) does not provide a free "
				"function!\n", hsm->description );
		return (PKI_ERR);
	}

	return (PKI_OK);
}


int HSM_init( HSM *hsm ) {

	if( !hsm ) return (PKI_ERR);

	/* Call the init function provided by the hsm itself */
	if( hsm->callbacks->init ) {
		return (hsm->callbacks->init ( hsm->driver, hsm->config ));
	} else {
		/* No init function is provided (not needed ??!?!) */
		PKI_log_debug("INFO, hsm (%s) does not provide an init "
				"function!\n", hsm->description );
	}

	return(PKI_OK);
}

PKI_KEYPAIR *PKI_KEYPAIR_new( int type, int bits, char *label,
					PKI_CRED *cred, HSM *hsm ) {

	if( bits < 1 ) return (NULL);

	if( hsm && hsm->callbacks->keypair_new ) {
		return hsm->callbacks->keypair_new( type, bits, 
					label, cred, hsm );
	} else {
		return HSM_OPENSSL_KEYPAIR_new( type, bits, label, cred, NULL );
	}

}

int PKI_KEYPAIR_free( PKI_KEYPAIR *key, HSM *hsm ) {

	if( !key ) return (PKI_OK);

	if( hsm && hsm->callbacks->keypair_free ) {
		return hsm->callbacks->keypair_free( key, hsm );
	} else {
		return HSM_OPENSSL_KEYPAIR_free( key, NULL );
	}
}

int PKI_X509_CERT_sign( PKI_X509_CERT *x, PKI_KEYPAIR *key,
		PKI_DIGEST_ALG *digest, HSM *hsm ) {

	return ( PKI_OBJ_sign( PKI_OBJTYPE_X509_CERT, x, key, digest, hsm ));
}

int PKI_X509_REQ_sign ( PKI_X509_REQ *x, PKI_KEYPAIR *key,
		PKI_DIGEST_ALG *digest, HSM *hsm ) {

	PKI_log_debug( "PKI_X509_REQ_sign() started" );
	return ( PKI_OBJ_sign( PKI_OBJTYPE_X509_REQ, x, key, digest, hsm ));
}

int PKI_X509_CRL_sign( PKI_X509_CRL *x, PKI_KEYPAIR *key,
		PKI_DIGEST_ALG *digest, HSM *hsm ) {

	return ( PKI_OBJ_sign( PKI_OBJTYPE_X509_CRL, x, key, digest, hsm ));
}

int PKI_sign ( PKI_OBJTYPE type, void *x, void *it,
		PKI_ALGOR *alg, PKI_BIT_STRING *bit,
		PKI_KEYPAIR *key, PKI_DIGEST_ALG *digest,
		HSM *hsm ) {

	int ret = PKI_ERR;

	PKI_log_debug( "PKI_sign() started" );

	if( !x || !key || !digest ) {
		if( !x ) PKI_log_debug("ERROR, missing data to sign!");
		if( !key ) PKI_log_debug("ERROR, missing KEY to sign with!");
		if( !digest ) PKI_log_debug("ERROR, missing DIGEST algorithm"
							" to sign with!");
		return (ret);
	}

	if( hsm && hsm->callbacks && hsm->callbacks->sign ) {
		PKI_log_debug( "HSM sign() callback called " );
		ret = hsm->callbacks->sign(type, x, it, alg, bit, key, 
								digest, hsm); 
	} else {

		PKI_log_debug( "SOFTWARE sign() callback called " );
		ERR_clear_error();
		PKI_log_debug( "DEBUG::sign()::%s:%d",__FILE__, __LINE__ );
		ret = HSM_OPENSSL_sign( type, x, it, alg, bit, 
							key, digest, NULL );
		PKI_log_debug( "DEBUG::sign()::%s:%d",__FILE__, __LINE__ );
		if( ret == PKI_ERR ) {
			PKI_log(PKI_LOG_ERR, "Signature: %s",
				ERR_error_string(ERR_get_error(), NULL));
		}
		PKI_log_debug( "DEBUG::sign()::%s:%d",__FILE__, __LINE__ );
	}

	PKI_log_debug( "PKI_sign() ended (%d)", ret );

	return (ret);
}

int PKI_verify ( PKI_OBJTYPE type, void *x, PKI_KEYPAIR *key, HSM *hsm ) {

	int ret = PKI_ERR;

	PKI_log_debug( "PKI_verify() started" );

	if( !x || !key ) {
		if( !x ) PKI_log_debug("ERROR, missing data to verify");
		if( !key ) PKI_log_debug("ERROR, missing keypair to verify with");
		return (ret);
	}

	if( hsm && hsm->callbacks && hsm->callbacks->verify ) {
		PKI_log_debug( "HSM verify() callback called " );
		ret = hsm->callbacks->verify(type, x, key, hsm); 
	} else {

		PKI_log_debug( "SOFTWARE verify() callback called " );
		ERR_clear_error();
		ret = HSM_OPENSSL_verify( type, x, key, NULL);
		if( ret == PKI_ERR ) {
			PKI_log(PKI_LOG_ERR, "Verify: %s",
				ERR_error_string(ERR_get_error(), NULL));
		}
	}

	PKI_log_debug( "PKI_verify() ended (%d)", ret );

	return (ret);
}



PKI_KEYPAIR *PKI_KEYPAIR_get ( char *url_s, PKI_CRED *cred, HSM *hsm ) {

	URL *url = NULL;

	if( !url_s ) return (NULL);

	if((url = URL_new( url_s )) == NULL ) {
		return(NULL);
	}

	return ( PKI_KEYPAIR_get_url ( url, cred, hsm ) );
}

PKI_KEYPAIR *PKI_KEYPAIR_get_url ( URL *url, PKI_CRED *cred, HSM *hsm ) {
	
        PKI_KEYPAIR *ret = NULL;

	if( !url ) return ( NULL );

	if( hsm  && hsm->callbacks && hsm->callbacks->keypair_get_url ) { 
		ret = hsm->callbacks->keypair_get_url( url, cred, hsm );
	} else {
		ERR_clear_error();
		ret = HSM_OPENSSL_KEYPAIR_get_url( url, cred, NULL );
		if( ret == NULL ) {
			ERR_print_errors_fp(stdout);
		}
	}

        return ( ret );
}

int PKI_KEYPAIR_export(PKI_KEYPAIR *x, char *label, int format, PKI_CRED *cred,
						char *url_string, HSM *hsm) {

	int ret = PKI_ERR;
	URL *url = NULL;
	
	/* We need a URL where to save it to... */
	if( !url_string ) return (PKI_ERR);

	/* We need at least one identifier - keypair or URL */
	if( !x && !label ) return ( PKI_ERR );

	if((url = URL_new( url_string )) == NULL ) {
		return( PKI_ERR );
	}

	return (PKI_KEYPAIR_export_url( x, label, format, cred, url, hsm ));

}

int PKI_KEYPAIR_export_url(PKI_KEYPAIR *x, char *label, int format, 
				PKI_CRED *cred, URL *url, HSM *hsm) {

	
	int ret = PKI_ERR;

	if( !url ) return (PKI_ERR);

	if( !x && !label ) return ( PKI_ERR );

	if( hsm && hsm->callbacks && hsm->callbacks->keypair_export_url ) {
		ret = hsm->callbacks->keypair_export_url ( x, label,
					format, url, cred, (void *) hsm );
	} else {
		ret = HSM_OPENSSL_KEYPAIR_export_url ( x, label,
					format, url, cred, (void *) hsm );
	}

	return ( ret );
}


