/* HSM Object Management Functions */

#include <strings.h>
#include <libpki/pki.h>
#include <libpki/hsm_st.h>

/* Callbacks for Software OpenSSL HSM */
HSM_CALLBACKS openssl_hsm_callbacks = {
		/* Init */
		HSM_OPENSSL_init,
		/* Free */
		HSM_OPENSSL_free,
		/* General Sign */
		HSM_OPENSSL_sign,
		/* General Verify */
		HSM_OPENSSL_verify,
		/* Key Generation */
		HSM_OPENSSL_KEYPAIR_new,
		/* Free Keypair Function */
		HSM_OPENSSL_KEYPAIR_free,
		/* Key Remove Function */
		NULL,
		/* Key Load Function */
		HSM_OPENSSL_KEYPAIR_get_url,
		/* Key Put Function */
		HSM_OPENSSL_KEYPAIR_export_url
};

/* Structure for PKI_TOKEN definition */
HSM openssl_hsm = {

        /* Version of the token */
        1,

        /* Description of the HSM */
        "OpenSSL Software HSM",

        /* Manufacturer */
        "OpenSSL Project",

        /* Pointer to the HSM config file and parsed structure*/
        NULL, 

        /* One of PKI_HSM_TYPE value */
        HSM_TYPE_SOFTWARE,

	/* URL for the ID of the driver, this is filled at load time */
	NULL,

	/* Pointer to the driver structure */
	NULL,

        /* Callbacks Structures */
	&openssl_hsm_callbacks
};

HSM *HSM_OPENSSL_new ( PKI_CONFIG *conf ) {
	HSM *hsm = NULL;

	hsm = (HSM *) PKI_Malloc ( sizeof( HSM ));
	memcpy( hsm, &openssl_hsm, sizeof( HSM));

	/* Not really needed! */
	hsm->callbacks = &openssl_hsm_callbacks;

	return( hsm );
}

int HSM_OPENSSL_free ( void *driver ) {

	if( driver == NULL ) return (PKI_OK);

	return (PKI_ERR);
}

int HSM_OPENSSL_init( void *driver, PKI_CONFIG *conf ) {

	/* No need for initialization of the software driver */
	return (PKI_OK);
}

/* General Signing function */
int HSM_OPENSSL_sign (PKI_OBJTYPE type, 
				void *x, 
				void *it_pp, 
				PKI_ALGOR *alg,
				PKI_STRING *bit,
				PKI_KEYPAIR *key, 
				PKI_DIGEST_ALG *digest, 
				void *driver ) {

	int ret = 0;
	ASN1_ITEM *it = NULL;

	if( !x || !key ) {
		PKI_log_debug("Missing required param for signature generation "
				"(Software HSM)");

		if( !x ) PKI_log_debug ( "Missing data to sign!");
		if( !key ) PKI_log_debug( "Missing Key to sign with!");

		return ( PKI_ERR );
	}

	if( !digest ) digest = PKI_DIGEST_ALG_SHA1;

	ERR_clear_error();

	switch ( type ) {
		case PKI_OBJTYPE_X509_REQ:
			ret = X509_REQ_sign( (X509_REQ *) x, 
				(EVP_PKEY *) key, (EVP_MD *) digest );
			break;
		case PKI_OBJTYPE_X509_CERT:
			ret = X509_sign( (X509 *) x, (EVP_PKEY *) key, 
				(EVP_MD *) digest );
			break;
		case PKI_OBJTYPE_X509_CRL:
			ret = X509_CRL_sign( (X509_CRL *) x, (EVP_PKEY *) key, 
				(EVP_MD *) digest );
			break;
		case PKI_OBJTYPE_PKCS7:
		case PKI_OBJTYPE_PKCS12:
		case PKI_OBJTYPE_PKI_MSG:
		case PKI_OBJTYPE_SCEP_MSG:
		case PKI_OBJTYPE_CMS_MSG:
			PKI_log_debug("HSM::DRIVER::OPENSSL::OBJ sign not "
					"supported for this type, yet!");
			ret = 0;
			break;
		default:
			if( !it_pp || !bit || !alg ) {
				PKI_log_debug("Missing required params to "
					"complete the generic signature");
				return ( PKI_ERR );
			}

			it = (ASN1_ITEM *) it_pp;
			ret = ASN1_item_sign(it, alg, NULL,
				bit, x, (EVP_PKEY *) key, (EVP_MD *) digest );
			break;
	}
			
	if( ret == 0 ) {
		PKI_log_debug("ERROR::Software::sign()::%s", 
				ERR_error_string(ERR_get_error(), NULL));
		return( PKI_ERR );
	}

	PKI_log_debug("Signature successful ( Software HSM )");

	return ( PKI_OK );
}

int HSM_OPENSSL_verify ( PKI_OBJTYPE type, void *x, 
					PKI_KEYPAIR *key, HSM *hsm ) {

        int ret = 0;
        EVP_PKEY *pp = NULL;
	X509_ALGOR *alg = NULL;

	void * sig = NULL;
	void * data = NULL;
	EVP_PKEY *pkey = NULL;

	ASN1_ITEM *it = NULL;

        if( !x || !key ) {
                PKI_log_debug("PKI_verify() - Missing resp or key!");
                return (PKI_ERR);
        }

        pp = (EVP_PKEY *) key;

        ERR_clear_error();

	switch ( type ) {
		case PKI_OBJTYPE_X509_REQ:
			it = (ASN1_ITEM *) ASN1_ITEM_rptr(X509_REQ_INFO);
			sig = ((X509_REQ *) x)->signature;
			data = ((X509_REQ *) x)->req_info;
			alg = ((X509_REQ *) x)->sig_alg;
			break;
		case PKI_OBJTYPE_X509_CERT:
			it = (ASN1_ITEM *) ASN1_ITEM_rptr(X509_CINF);
			sig = ((X509 *)x)->signature;
			data = ((X509 *)x)->cert_info;
			alg = ((X509 *)x)->sig_alg;
			break;
		case PKI_OBJTYPE_X509_CRL:
			it = (ASN1_ITEM *) ASN1_ITEM_rptr(X509_CRL_INFO);
			sig = ((X509_CRL *)x)->signature;
			data = ((X509_CRL *)x)->crl;
			alg = ((X509_CRL *)x)->sig_alg;
			break;
		case PKI_OBJTYPE_PRQP_REQ:
			it = (ASN1_ITEM *) ASN1_ITEM_rptr(TBS_REQ_DATA);
			sig = ((PRQP_REQ *)x)->prqpSignature->signature;
			data = ((PRQP_REQ *)x)->requestData;
			alg = ((PRQP_REQ *)x)->prqpSignature->signatureAlgorithm;
			break;
		case PKI_OBJTYPE_PRQP_RESP:
			it = (ASN1_ITEM *) ASN1_ITEM_rptr(TBS_RESP_DATA);
			sig = ((PRQP_RESP *)x)->prqpSignature->signature;
			data = ((PRQP_RESP *)x)->respData;
			alg = ((PRQP_RESP *)x)->prqpSignature->signatureAlgorithm;
			break;
		case PKI_OBJTYPE_PKCS7:
		case PKI_OBJTYPE_PKCS12:
		case PKI_OBJTYPE_PKI_MSG:
		case PKI_OBJTYPE_SCEP_MSG:
		case PKI_OBJTYPE_CMS_MSG:
			PKI_log_debug("HSM::DRIVER::OPENSSL::OBJ verify not "
					"supported for this type, yet!");
			return( PKI_ERR );
			break;
		default:
			return ( PKI_ERR );
	}

	if( !it || !alg || !sig || !data || !pp ) return ( PKI_ERR );

        if(( ret = ASN1_item_verify((const ASN1_ITEM *)it, alg, sig, data, 
							pp )) == 0 ) {

                PKI_log_debug( "PKI_verify() - [%d] ERROR:%s", ret,
                        ERR_error_string(ERR_get_error(), NULL ));

                return ( PKI_ERR );
        }

        PKI_log_debug( "PKI_verify() - OK");

        return ( PKI_OK );

}
