/* openssl/pki_pkey.c */

#include <libpki/pki.h>

/* Internal usage only - we want to keep the lib abstract */
#ifndef _LIBPKI_HSM_OPENSSL_PKEY_H
#define _LIBPKI_HSM_OPENSSL_PKEY_H

#define PKI_RSA_KEY	RSA
#define PKI_DSA_KEY	DSA

#ifdef ENABLE_ECDSA
#define PKI_EC_KEY	EC_KEY
#endif

#define PKI_RSA_KEY_MIN_SIZE		512
#define PKI_DSA_KEY_MIN_SIZE		512
#define PKI_EC_KEY_MIN_SIZE		56

PKI_RSA_KEY * _pki_rsakey_new( int bits );
PKI_DSA_KEY * _pki_dsakey_new( int bits );
#ifdef ENABLE_ECDSA
PKI_EC_KEY * _pki_ecdsakey_new( int bits);
#else
void * _pki_ecdsakey_new( int bits );
#endif

int _pki_rand_init( void );

/* End of _LIBPKI_INTERNAL_PKEY_H */
#endif

int _pki_rand_seed( void ) {
	unsigned char seed[20];

	if (!RAND_pseudo_bytes(seed, 20)) {
		return 0;
	}
	RAND_seed(seed, sizeof seed);

	return(1);
}

PKI_RSA_KEY * _pki_rsakey_new( int bits ) {
	BIGNUM *bn = NULL;
	unsigned long esp = 0x10001;
	PKI_RSA_KEY *rsa = NULL;

	if( bits < PKI_RSA_KEY_MIN_SIZE )
		return NULL;

	// if( (rsa = RSA_generate_key(bits, bn, NULL, NULL)) == NULL ) {
	if( (rsa = RSA_generate_key(bits, esp, NULL, NULL)) == NULL ) {
		/* Error */
		// RSA_free( rsa );
		BN_free( bn );
		PKI_log_debug("ERROR, can not generate RSA key");
		return NULL;
	}

	/* Free memory */
	// if( bn ) BN_free (bn);

	/* Let's return the RSA_KEY infrastructure */
	return (rsa);
}

PKI_DSA_KEY * _pki_dsakey_new( int bits ) {
	PKI_DSA_KEY *k = NULL;
	unsigned char seed[20];

	if(bits < PKI_DSA_KEY_MIN_SIZE )  return NULL;

	if (!RAND_pseudo_bytes(seed, 20)) {
		/* Not enought rand ? */
		return NULL;
	}
	/*
	if((k = DSA_new()) == NULL ) {
		fprintf(stderr, "pla1\n");
		return NULL;
	}
	*/

	// if(!DSA_generate_parameters_ex( k, bits,
	if((k = DSA_generate_parameters( bits,
				seed, 20, NULL, NULL, NULL, NULL)) == NULL ) {
		if( k ) DSA_free( k );
		PKI_log_debug("ERROR, can not generate DSA params");
		return NULL;
	}

	return( k );
}

#ifdef ENABLE_ECDSA
PKI_EC_KEY * _pki_ecdsakey_new( int bits ) {
	/* ECDSA is a little more complicated than the other
	   schemes as it involves a group of functions. As the
	   purpose of this library is to provide a very hi-level
	   easy to use library, we will provide some hardwired
	   parameters.
	*/
	PKI_EC_KEY *k = NULL;
	EC_builtin_curve *curves = NULL;
	EC_GROUP *group = NULL;
	int num_curves = 0;
	int i = 0;
	unsigned int rnd_num = 0;
	int nid = 0;
	int degree = 0;
	int n_start = 0;

	if( bits < PKI_EC_KEY_MIN_SIZE ) return NULL;

	/* Get the number of availabe ECDSA curves in OpenSSL */
        if((num_curves = EC_get_builtin_curves(NULL, 0)) < 1 ) {
		/* No curves available! */
		return NULL;
	}

	/* Alloc the needed memory */
        curves = OPENSSL_malloc(sizeof(EC_builtin_curve) * num_curves);

	/* Check for memory allocation */
	if (curves == NULL) return NULL;

	/* Get the builtin curves */
	if (!EC_get_builtin_curves(curves, num_curves)) {
		if( curves ) free (curves);
		return NULL;
	}

	/* We completely change behavior - we adopt one of the two
         * curves suggested by NIST. In particular:
         * - NID_secp384r1
         * - NID_secp521r1
         * For today (2008) usage, the first curve + SHA256 seems to be
         * the best approach
         */

	if( bits <= 112 ) {
		nid = NID_secp112r1;
	} else if( bits <= 128 ) {
		nid = NID_secp128r1;
	} else if( bits <= 160 ) {
		nid = NID_secp160r1;
	} else if( bits <= 192 ) {
		nid = NID_X9_62_prime192v1;
	} else if( bits <= 224 ) {
		nid = NID_secp224r1;
	} else if( bits <= 256 ) {
		nid = NID_X9_62_prime256v1;
	} else if( bits <= 384 ) {
		nid = NID_secp384r1;
	} else {
		/* This is the only curve that supports SHA512! */
		nid = NID_secp521r1;
	};

	// // Get a random curve among the available ones
	// do {
	// 	RAND_pseudo_bytes( (unsigned char *) &rnd_num, sizeof(rnd_num));
	// 	i = rnd_num % num_curves;
	// 	nid = curves[i].nid;
	// } while ( nid == NID_ipsec4 );

	// n_start = i;
	// do {
	// 	// We have to match the strenght required for the
	// 	// private key, let' cycle until we have curves
	//
	//	nid = curves[i].nid;
	//

	/* Initialize the key */
	if ((k = EC_KEY_new()) == NULL) {
		if( curves) free (curves);
		return NULL;
	}

	if((group = EC_GROUP_new_by_curve_name(nid)) == NULL ) {
		if ( k ) EC_KEY_free( k );
		if ( curves ) free (curves);
		return NULL;
	};

	/* Assign the group to the key */
	if (EC_KEY_set_group(k, group) == 0) {
		if ( group ) EC_GROUP_free( group );
		if ( k ) EC_KEY_free( k );
		if ( curves ) free (curves);
		return NULL;
	}

	/* We do not need it now, let's free the group */
	if ( group ) EC_GROUP_free( group );
	group = NULL;

	degree = EC_GROUP_get_degree(EC_KEY_get0_group(k));

	if( degree < bits ) {
		/* Fix the problem, let's get the right bits */
		bits = degree;
	}

	//	// Let's cycle through all the available curves
	//	// until we find one that matches (if any)
	//	i = (i + 1 ) % num_curves;
	//
	// } while ( (degree < bits ) && (i != n_start) );

	/* Now generate the key */
	if (!EC_KEY_generate_key(k)) {
		if( curves ) free (curves);
		if( k ) EC_KEY_free (k);
		PKI_log_debug("ERROR, generating ecdsa key");
		return NULL;
	}

	/* Verify the Key to be ok */
	if (!EC_KEY_check_key(k)) {
		if( curves ) free (curves);
		if( k ) EC_KEY_free (k);
		PKI_log(PKI_LOG_ERR, "Can not verify ecdsa key");
		return NULL;
	}

	return ( k );
}

#else /* EVP_PKEY_EC */

void * _pki_ecdsakey_new( int bits ) {
	return ( NULL );
}

#endif


PKI_KEYPAIR *HSM_OPENSSL_KEYPAIR_new( int type, int bits, 
				char *label, PKI_CRED *cred, void *driver ) {
	PKI_KEYPAIR *ret = NULL;
	PKI_RSA_KEY *rsa = NULL;
	PKI_DSA_KEY *dsa = NULL;
#ifdef ENABLE_ECDSA
	PKI_EC_KEY *ec = NULL;
#endif

	if((ret = (PKI_KEYPAIR *) EVP_PKEY_new()) == NULL ) {
		return NULL;
	}

	if( _pki_rand_seed() == 0 ) {
		/* Probably low level of randomization available */
		PKI_log_debug("WARNING, low rand available!");
	}

	switch (type) {
		case PKI_SCHEME_RSA:
			if((rsa = _pki_rsakey_new( bits )) == NULL ) {
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return NULL;
			};
			if(!EVP_PKEY_assign_RSA( (EVP_PKEY *) ret, rsa)) {
				if( rsa ) RSA_free( rsa );
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return NULL;
			};
			break;
		case PKI_SCHEME_DSA:
			if((dsa = _pki_dsakey_new( bits )) == NULL ) {
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return(NULL);
			};
			if (!DSA_generate_key( dsa )) {
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return NULL;
			}
                        if (!EVP_PKEY_assign_DSA( (EVP_PKEY *) ret, dsa)) {
				if( dsa ) DSA_free ( dsa );
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return NULL;
			}
                        dsa=NULL;
			break;
#ifdef ENABLE_ECDSA
		case PKI_SCHEME_ECDSA:
			if((ec = _pki_ecdsakey_new( bits )) == NULL ) {
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return(NULL);
			};
			if (!EC_KEY_generate_key(ec)) {
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return NULL;
			}
                        if (!EVP_PKEY_assign_EC_KEY( (EVP_PKEY *) ret, ec)) {
				if( ec ) EC_KEY_free ( ec );
				if( ret ) PKI_KEYPAIR_free( ret, NULL );
				return NULL;
			}
			ec=NULL;
			break;
#endif
		default:
			/* No recognized scheme */
			if( ret ) PKI_KEYPAIR_free( ret, NULL );
			return NULL;
	}

	/* Let's return the PKEY infrastructure */
	return ( ret );
}

/* Key Free function */
int HSM_OPENSSL_KEYPAIR_free ( PKI_KEYPAIR *pkey, void *driver ) {
	if( !pkey) return (PKI_ERR);

	EVP_PKEY_free( (EVP_PKEY *) pkey);

	return(PKI_OK);

}

/* Key Pair removal function */
int HSM_OPENSSL_KEYPAIR_remove ( PKI_KEYPAIR *key, char *label, void *driver ) {

	PKI_log_debug( "HSM_OPENSSL_KEYPAIR_get() not implemented!\n");
	return (PKI_ERR);
}

PKI_KEYPAIR *HSM_OPENSSL_KEYPAIR_get_url( URL *url,
					PKI_CRED *cred, void *driver ) {

        PKI_KEYPAIR *ret = NULL;
	PKI_MEM_STACK *data = NULL;
	PKI_MEM *obj = NULL;
	BIO *bio = NULL;
	char *pp = NULL;

	if(!url) return NULL;

	if( cred && cred->password ) pp = (char *) cred->password;

	if( url->proto == URI_PROTO_ID ) {
		/* We do not know how to load a private key, because it is
		   handled directly by the HSM */
		PKI_log_debug("ERROR, ID type of URI used in software HSM!");
		return ( NULL );
	}

	if(( data = URL_get_data_url( url, 0 )) == NULL ) {
		PKI_log_debug("ERROR, can not get key data from URL");
		return(NULL);
	}

	if(( obj = PKI_STACK_MEM_pop( data )) == NULL ) {
		if( data ) PKI_STACK_MEM_free ( data );
		PKI_log_debug("ERROR, Memory Allocation (%s:%d)",
					__FILE__, __LINE__ );
		return ( NULL );
	}

	if(( bio = BIO_new_mem_buf( obj->data, obj->size )) == NULL ) {
		if( data ) PKI_STACK_MEM_free (data);
		if( obj ) PKI_MEM_free (obj);
		PKI_log_debug("ERROR, Memory Allocation (%s:%d)",
							__FILE__, __LINE__ );
		return ( NULL );
	}

	if((ret = (PKI_KEYPAIR *) PEM_read_bio_PrivateKey(bio, NULL, 
						NULL, pp )) == NULL) {

		PKI_log_debug("INFO, Can not load Key (PEM format)" );

		ERR_print_errors_fp(stderr);
		ERR_clear_error();

		/* Not PEM formatted (?) */
		if(BIO_reset(bio) == 1 ) {
			ret = (PKI_KEYPAIR *) d2i_PrivateKey_bio(bio, NULL);
			if(!ret ) {
				PKI_log_debug("ERROR, Can not load Key (not "
						"PEM nor DER format)" );
				// ERR_print_errors_fp(stderr);
			}
		}
	}

	if( data ) PKI_STACK_MEM_free ( data );
	if( obj ) PKI_MEM_free (obj);

        return ret;
}

int HSM_OPENSSL_KEYPAIR_export_url( PKI_KEYPAIR *key, char *label, 
			int format, URL *url, PKI_CRED *cred, void *driver ) {

	if( !url ) return ( PKI_ERR );

	switch ( url->proto ) {
		case URI_PROTO_FILE:
			return PKI_OPENSSL_KEYPAIR_export_file( key, format,
					url, cred );
			break;
		default:
			return ( PKI_ERR );
	}

	return ( PKI_OK );
}

int PKI_OPENSSL_KEYPAIR_export_file( PKI_KEYPAIR *key, int format, URL *url,
						PKI_CRED *cred ) {

	BIO *out = NULL;
	int ret = 1;

	if( !key || !url ) return 0;

	switch( format ) {
		case PKI_FORMAT_PEM:
		case PKI_FORMAT_ASN1:
			break;
		default:
			PKI_log_debug( "%s:%d format not recognized (%d)\n",
					__FILE__, __LINE__, format );
			return(PKI_ERR);
	}

	if((out = BIO_new(BIO_s_file())) == NULL ) {
		PKI_log_debug("%s:%d error\n", __FILE__, __LINE__ );
		return (PKI_ERR);
	}

	if( strncmp_nocase( url->addr, "-", 1) == 0 ) {
		 BIO_set_fp(out,stdout,BIO_NOCLOSE);
	} else {
		ret = (int ) BIO_write_filename( out, url->addr);
	}

	switch( format ) {
		case PKI_FORMAT_PEM:
			ret = PEM_write_bio_PrivateKey( out, (EVP_PKEY *) key,
					NULL, NULL, 0, NULL, NULL );
			break;
		case PKI_FORMAT_ASN1:
			ret = i2d_PrivateKey_bio( out, (EVP_PKEY *) key);
			break;
		default:
			PKI_log_debug("%s:%d error\n", __FILE__, __LINE__ );
	}

	if( out ) BIO_free_all (out);

	return(ret);
}

/*
PKI_KEYPAIR * HSM_OPENSSL_KEYPAIR_get_url ( URL *url, PKI_CRED *cred,
							void * driver ) {

	printf("%s:%d::DEBUG::Not Implemented!\n", __FILE__, __LINE__ );

	return ( NULL );
}

*/
