/* PKI_X509 I/O management */

#include <libpki/pki.h>

/*! \brief Retrieve a certificate from a URL
 *
 * Downloads a certificate from a given URL (file://, http://, ldap://...)
 * in (char *) format.
 * The returned data is of type PKI_X509_CERT in case of success or NULL if
 * any error occurred. If multiple objects are returned from the URL, only
 * the first one is returned. Use PKI_X509_CERT_STACK_get() function
 * to retrieve a PKI_X509_CERT_STACK * object.
 *
 */

PKI_X509_CERT *PKI_X509_CERT_get ( char *url_s ) {

	PKI_X509_CERT * ret = NULL;
	URL *url = NULL;

	if( !url_s ) return (NULL);

	if((url = URL_new( url_s )) == NULL ) {
		return (NULL);
	}

	if((ret = PKI_X509_CERT_get_url ( url )) == NULL ) {
		return (NULL);
	}

	return( ret );

}

/*! \brief Retrieve a certificate from a URL pointer.
 *
 * Downloads a certificate from a given URL (file://, http://, ldap://...)
 * in (URL *) format. To generate a URL * from a char * use URL_new().
 * The returned data is of type PKI_X509_CERT in case of success or NULL if
 * any error occurred. If multiple objects are returned from the URL, only
 * the first one is returned. Use PKI_X509_CERT_STACK_get_url() function
 * to retrieve a PKI_X509_CERT_STACK * object.
 *
 */

PKI_X509_CERT *PKI_X509_CERT_get_url ( URL *url ) {

	PKI_X509_CERT_STACK *sk = NULL;
	PKI_X509_CERT * ret = NULL;

	if( !url ) return (NULL);

	if((sk = PKI_X509_CERT_STACK_get_url ( url )) == NULL ) {
		return (NULL);
	}

	if( PKI_STACK_X509_CERT_elements( sk ) >= 1 ) {
		ret = PKI_STACK_X509_CERT_pop( sk );
	}

	if( sk ) PKI_STACK_X509_CERT_free ( sk );

	return ( ret );

}

/*! \brief Retrieve a stack of certificates from a URL (char *).
 *
 * Downloads a stack of certificates from a given URL (file://, http://,
 * ldap://...) passed as a (char *).
 *
 * The returned data is a pointer to a PKI_X509_CERT_STACK data structure
 * in case of success or NULL if any error occurred.
 * If only the first object is required from the URL, use the 
 * PKI_X509_CERT_get_url() function instead.
 *
 */

PKI_X509_CERT_STACK *PKI_X509_CERT_STACK_get ( char *url_s ) {

	URL *url = NULL;

	if( !url_s ) return (NULL);

	if((url = URL_new( url_s )) == NULL ) {
		return(NULL);
	}

	return ( PKI_X509_CERT_STACK_get_url ( url ) );
}

/*! \brief Retrieve a stack of certificates from a URL (URL *) pointer.
 *
 * Downloads a stack of certificates from a given URL (file://, http://,
 * ldap://...) passed as a (URL *).  To generate a (URL *) from a (char *)
 * use URL_new().
 *
 * The returned data is a pointer to a PKI_X509_CERT_STACK data structure
 * in case of success or NULL if any error occurred.
 * If only the first object is required from the URL, use the 
 * PKI_X509_CERT_get_url() function instead.
 *
 */

PKI_X509_CERT_STACK *PKI_X509_CERT_STACK_get_url ( URL *url ) {
	
        PKI_X509_CERT_STACK *ret = NULL;
        PKI_X509_CERT_STACK *tmp_x_sk = NULL;
	PKI_MEM_STACK *mem_sk = NULL;
	PKI_X509_CERT *x = NULL;

	int i;

	if(!url) return NULL;

	if((mem_sk = URL_get_data_url ( url, 0 )) == NULL ) {
		return(NULL);
	}

	if((ret = PKI_STACK_X509_CERT_new()) == NULL ) {
		return(NULL);
	}

	for( i = 0; i < PKI_STACK_MEM_elements( mem_sk ); i++ ) {
		PKI_MEM *n = NULL;

		if(( n = PKI_STACK_MEM_get_num( mem_sk, i )) == NULL ) {
			break;
		}

		if((tmp_x_sk = PKI_X509_CERT_STACK_get_mem( n )) != NULL ) {
			while ( (x = PKI_STACK_X509_CERT_pop( tmp_x_sk ))
								 != NULL ) {
				PKI_STACK_X509_CERT_push( ret, x );	
			}
			PKI_STACK_X509_CERT_free ( tmp_x_sk );
		}
	}

	if( mem_sk ) PKI_STACK_MEM_free_all ( mem_sk );

        return ( ret );
}

/*
char * memstr( char *str, char *tok, int size ) {
        int i, j, tok_size, found;

        if( !str || !tok || size < 1 ) return (NULL);

        tok_size = strlen( tok );

        found = 0;
        for( i = 0; i < size; i++ ) {
                found = 1;
                for( j = 0; j < tok_size; j++ ) {
                        if( str[i+j] != tok[j] ) {
                                found = 0;
                                break;
                        }
                }
                if ( found == 1 ) {
                        return ( &str[i] );
                }
        }

        return( NULL );
}
*/

char * memstr( char *str, char *tok, ssize_t size ) {
	int i, j, tok_size, found;

	if( !str || !tok || size < 1 ) return (NULL);

	tok_size = strlen( tok );

	found = 0;
	for( i = 0; i < size; i++ ) {
		found = 1;
		for( j=0; j < tok_size; j++ ) {
			if( str[i+j] != tok[j] ) {
				found = 0;
				break;
			}
		}
		if ( found == 1 ) return ( &str[i] );
	}

	return( NULL );
}


PKI_X509_CERT_STACK *PKI_X509_CERT_STACK_get_mem( PKI_MEM *mem ) {

	PKI_X509_CERT * x = NULL;
	PKI_X509_CERT_STACK * sk = NULL;
	BUF_MEM *p = NULL;
	BIO *membio = NULL;

	char *temp = NULL;
	int cont = 1;
	long curr = 0;

	int mem_size = 0;
	char * mem_buf = NULL;

	if( !mem || mem->size <= 0 ) return (NULL);

	if((temp = strstr(mem->data, PKI_X509_CERT_BEGIN_ARMOUR )) == NULL ) {
		mem_buf = mem->data;
		mem_size = mem->size;
	} else {
		mem_buf = temp;
		mem_size = mem->data + mem->size - temp;
	}

	if((membio = BIO_new_mem_buf( mem_buf, mem_size )) == NULL ) {
		return( NULL );
	}

	if(BIO_set_close(membio, BIO_NOCLOSE) != 1 ) {
		BIO_free (membio);
		return(NULL);
	}

	if((sk = PKI_STACK_X509_CERT_new()) == NULL ) {
		BIO_free( membio );
		return(NULL);
	}

	cont = 1;
	p = (BUF_MEM *) membio->ptr;
	curr = p->length;
	while ((cont == 1) && (p->length > 0)) {

		curr = p->length;

		/* Try PEM format */
        	if(( x = (PKI_X509_CERT *) 
			PEM_read_bio_X509(membio,NULL,NULL,NULL)) == NULL){

			/* Resetting the BIO to previous pointer */
			p->length = curr;
			p->data -= curr;

			/* Is it DER encoded (???) */
			if((x=(PKI_X509_CERT *) d2i_X509_bio(membio,NULL)) == NULL ) {
				/* Format is not recognized! */
				cont = 0;
			} else {
				/* Certificate is in DER format */
				PKI_STACK_X509_CERT_push( sk, x );
			}
		} else {
			/* Certificate is in PEM format */
			PKI_STACK_X509_CERT_push( sk, x );

		}
	}

	if(membio) BIO_free( membio );
	return( sk );
}


int PKI_X509_CERT_write ( PKI_X509_CERT *x, int format, char *url_string ) {

	PKI_X509_CERT_STACK *sk = NULL;
	int ret = PKI_OK;

	if( !x || !url_string ) return (PKI_ERR);

	if(( sk = PKI_STACK_X509_CERT_new()) == NULL ) {
		return( PKI_ERR );
	}

	if( PKI_STACK_X509_CERT_push( sk, x ) == PKI_ERR ) {
		ret = PKI_ERR;
	} else {
		ret = PKI_X509_CERT_STACK_write( sk, format, url_string);
	}

	if( sk ) {
		// PKI_X509_CERT *tmp_x = NULL;

		while((x = PKI_STACK_X509_CERT_pop( sk )) != NULL ) {
			PKI_STACK_X509_CERT_free( sk );
		}
	}

	return (ret);

}

int PKI_X509_CERT_STACK_write (PKI_X509_CERT_STACK *sk, int format, 
							char *url_string) {

	URL *url = NULL;

	if( !sk || !url_string ) return (PKI_ERR);

	if((url = URL_new (url_string)) == NULL ) {
		return (PKI_ERR);
	}

	return (PKI_X509_CERT_STACK_write_url( sk, format, url ));
	
}


int PKI_X509_CERT_STACK_write_url (PKI_X509_CERT_STACK *sk, int format, 
								URL *url ) {

	if( !sk || !url ) return (PKI_ERR);

	switch( url->proto ) {
		case URI_PROTO_FILE:
			return PKI_X509_CERT_STACK_write_file( sk, format, 
							url->addr);
			break;
	}

	return( PKI_ERR );
}


int PKI_X509_CERT_chain_write(PKI_X509_CERT_STACK *sk,int format,char *url_s) {

	URL *url = NULL;
	int ret = PKI_OK;

	if( !sk || !url_s ) return (PKI_ERR);

	if((url = URL_new (url_s)) == NULL ) {
		return (PKI_ERR);
	}

	ret = PKI_X509_CERT_chain_write_url( sk, format, url );

	if( url ) URL_free (url);

	return(ret);
}

int PKI_X509_CERT_chain_write_url(PKI_X509_CERT_STACK *sk,int format,URL *url) {

	if( !sk || !url ) return (PKI_ERR);

	switch( url->proto ) {
		case URI_PROTO_FILE:
			return PKI_X509_CERT_STACK_write_file( sk, format, 
								url->addr);
			break;
	}

	return( PKI_ERR );
}


/* 
 * File Operations for Certificates
 */

/*
PKI_X509_CERT_STACK *PKI_X509_CERT_STACK_get_file ( URL *url ) {

        PKI_X509_CERT *x = NULL;
        BIO *in = NULL;
        PKI_X509_CERT_STACK *x_sk = NULL;

	// Check the pointer
	if( !url || !url->addr ) return NULL;

	// Open the file in addr
        if ((in=BIO_new_file( url->addr, "r")) == NULL) {
                return(NULL);
        }

	// Initialize the structure
        if((x_sk = PKI_STACK_X509_CERT_new()) == NULL ) {
                return(NULL);
        }

	// Try PEM FORMAT
	while(( x = (PKI_X509_CERT *) PEM_read_bio_X509(in, NULL, NULL, NULL)) 
								== NULL ) {
		PKI_STACK_X509_CERT_push(x_sk, x);
	}

	if( PKI_STACK_X509_CERT_elements( x_sk ) < 1 ) {
		// No PEM formatted files have been found
		while(( x = (PKI_X509_CERT *) d2i_X509_bio(in, NULL)) == NULL ) {
			// DER formatted Certificate found
			PKI_STACK_X509_CERT_push(x_sk, x);
		}
	}

	if( PKI_STACK_X509_CERT_elements(x_sk) == 0 ) {
		PKI_STACK_X509_CERT_free (x_sk );
		return NULL;
	}
        return x_sk;
}
*/

int PKI_X509_CERT_write_file ( PKI_X509_CERT *x, int format, char *file ) {

	BIO *out = NULL;
	int ret = PKI_OK;

	if( !x ) return (PKI_ERR);

	if((out = BIO_new(BIO_s_file())) == NULL ) {
		/* Memory error (?) */
		printf("%s:%d error\n", __FILE__, __LINE__ );
		return (PKI_ERR);
	}

	if( !file ) {
		/* standard output selected */
		 BIO_set_fp(out,stdout,BIO_NOCLOSE);
	} else {
		ret = BIO_write_filename( out, file);
		if( ret == 0 ) {
			if( out ) BIO_free_all (out);
			return (PKI_ERR);
		}
	}

	switch( format ) {
		case PKI_FORMAT_PEM:
			ret = PEM_write_bio_X509( out, (X509 *) x);
			break;
		case PKI_FORMAT_ASN1:
			ret = i2d_X509_bio( out, (X509 *) x);
			break;
		default:
			/* Format not recognized ! */
			printf("%s:%d error\n", __FILE__, __LINE__ );
			return(PKI_ERR);
	}

	if( out ) BIO_free_all (out);

	if( ret == 0 ) 
		return (PKI_ERR);
	else
		return(PKI_OK);

}

int PKI_X509_CERT_STACK_write_file(PKI_X509_CERT_STACK *sk, int format, 
								char *file){

	BIO *out = NULL;
	int ret = PKI_OK;
	int i = 0;

	if( !sk ) return (PKI_ERR);

	if((out = BIO_new(BIO_s_file())) == NULL ) {
		/* Memory error (?) */
		printf("%s:%d error\n", __FILE__, __LINE__ );
		return (PKI_ERR);
	}

	if( !file ) {
		/* standard output selected */
		 BIO_set_fp(out,stdout,BIO_NOCLOSE);
	} else {
		ret = BIO_write_filename( out, file);
		if( ret == 0 ) {
			if( out ) BIO_free_all (out);
			return (PKI_ERR);
		}
	}

	for( i = 0; i < PKI_STACK_X509_CERT_elements ( sk ); i++ ) {
		PKI_X509_CERT *curr_cert = NULL;

		if((curr_cert = PKI_STACK_X509_CERT_get_num( sk, i ))
							== NULL ) {
			break;
		}

		switch( format ) {
			case PKI_FORMAT_PEM:
				ret = PEM_write_bio_X509( out, 
							(X509 *) curr_cert);
				break;
			case PKI_FORMAT_ASN1:
				ret = i2d_X509_bio( out, (X509 *) curr_cert);
				break;
			default:
				/* Format not recognized ! */
				printf("%s:%d error\n", __FILE__, __LINE__ );
				return(PKI_ERR);
		}
	}

	if( out ) BIO_free_all (out);

	if( ret == 0 ) 
		return (PKI_ERR);
	else
		return(PKI_OK);

}
