/* OpenCA libpki package
* (c) 2000-2007 by Massimiliano Pala and OpenCA Group
* All Rights Reserved
*
* ===================================================================
* Released under OpenCA LICENSE
*/

#include <libpki/pki.h>
#include <libpki/net/url.h>
#include <libpki/net/http.h>

int parse_http_headers_mem ( PKI_MEM *pki_mem ) {

        int head_lines = 0;
        char linebuf[1024];
	int read_chars = 0;

	BIO *in = NULL;

	if( (in = BIO_new( BIO_s_mem())) == NULL ) {
                return( -1 );
        }

        head_lines = 0;
        do {
                int read_code = 0;

                memset(linebuf, '\x0', sizeof linebuf);
                if((read_code = BIO_gets(in, linebuf, sizeof(linebuf))) < 0) {
                        PKI_log( PKI_LOG_ERR, "HTTP, Error (%d) retrieving "
					"data (head_lines=%d)", 
                                                read_code, head_lines );
                        return (-1);
                }

                /* Let's check if the first line has good code */
                if ( ( head_lines == 0 ) &&
                        ( strncmp_nocase( &(linebuf[9]), "200", 3) != 0 )) {
                                /* Got an error - probably not found (?) */
                                PKI_log( PKI_LOG_ERR, "HTTP, Error retrieving "
					" data (%s)", linebuf );
                                return ( -1 );
                } else {
                        head_lines++;
                }
		read_chars += strlen(linebuf);
        }
        while( (linebuf[0] != '\r') && (linebuf[0] != '\n') );

	BIO_free( in );

        return (read_chars);

}

PKI_MEM *http_get_data_sock ( int fd, ssize_t max_size ) {

        PKI_MEM *buf = NULL;
        unsigned long fullsize = 0;
        unsigned long newsize  = 0;

        if( fd < 1 ) return NULL;

        /* get subject name from bio using recommended OpenSSL template */
        buf = PKI_MEM_new_null();
        for (;;) {
                if ((buf == NULL) || (!PKI_MEM_grow( buf, fullsize+512))) {
                        return NULL;
                }

                newsize   = read ( fd, &(buf->data[fullsize]), 512);
                fullsize += newsize;
                if (newsize == 0) break; 

                if (newsize < 0) {
                        /* Error Reading from buf */
                        PKI_MEM_free( buf );
                        return NULL;
                }

                if( (max_size > 0) && (fullsize > max_size)) {
                        /* Max Reading size exceeded */
                        PKI_log( PKI_LOG_ERR, "HTTP, Max read size exceeded "
                                " [ %d ]", max_size );
                        PKI_MEM_free( buf );
                        return NULL;
                }
        }
        /* Make sure the buf ends with a \x0 value */
        buf->data[fullsize] = '\x0';

        return buf;
}


PKI_MEM_STACK *URL_get_data_http_url ( URL *url, ssize_t max_size ) {

	void *ctxt;
	unsigned char *buf = NULL;
	int len = 0;
	uint32_t tot_len = 0;
	PKI_MEM *obj = NULL;
	PKI_MEM_STACK *ret = NULL;
    
	ctxt = xmlNanoHTTPOpen(url->url_s, NULL);
	if (ctxt == NULL) {
		return(NULL);
	}

	obj = PKI_MEM_new_null();
	if(( buf = (unsigned char *) PKI_Malloc ( LIBPKI_HTTP_BUF_SIZE )) 
								== NULL) {
		xmlNanoHTTPClose(ctxt);
		return( NULL );
	}

	tot_len = 0;
	while ((len = xmlNanoHTTPRead(ctxt, buf, sizeof(buf))) > 0) {
		PKI_MEM_add( obj, buf, len );
		tot_len += len;
	}

	xmlNanoHTTPClose(ctxt);
	PKI_Free( buf );

	if((ret = PKI_STACK_MEM_new()) != NULL ) {
		PKI_STACK_push( ret, obj );
	}

	return( ret );
}

PKI_MEM_STACK *URL_get_data_http ( char *url_s, ssize_t max_size ) {
	URL *url = NULL;

	if( !url_s ) return (NULL);

	if((url = URL_new( url_s )) == NULL ) {
		return NULL;
	}

	if( url->proto != URI_PROTO_HTTP ) {
		if( url ) URL_free ( url );
		return( NULL );
	}

	return URL_get_data_http_url ( url, max_size );
}

PKI_MEM_STACK * HTTP_post_data ( char *url_s, char *data, ssize_t size, 
						char *content_type ) {

	URL *url = NULL;
	PKI_MEM_STACK *ret = NULL;

	if( !url_s || !data || !content_type ) {
		/* ERROR: All data are strictly required! */
		return(NULL);
	}

	if((url = URL_new(url_s)) == NULL ) {
		/* Error in creating the URL structure */
		return(NULL);
	}

	ret = URL_HTTP_post_data ( url, data, size, content_type );

	if ( url ) URL_free (url);

	return(ret);
}


PKI_MEM_STACK * URL_HTTP_post_data ( URL *url, char *data, ssize_t size, 
						char *content_type ) {

	int sock;
	char *headers = "POST %s HTTP/1.1\r\nContent-Type: %s\r\nContent-Length: %d\r\nHost: %s\r\n\r\n";
	char buffer[2048];
	int rv;
	PKI_MEM_STACK *ret = NULL;
	PKI_MEM *buf = NULL;
	PKI_MEM *mm = NULL;

	if( !url || !data || !content_type ) {
		/* ERROR: All data are strictly required! */
		return(NULL);
	}

	snprintf( buffer, sizeof(buffer), headers, url->path, 
					content_type, size, url->addr );

	sock = inet_connect( url );
	if( sock < 1 ) {
		/* Can not connect to the server */
		return(NULL);
	}

	if((rv = _Write( sock, buffer, strlen( buffer ))) < 1 ) {
		inet_close( sock );
		return( NULL );
	}

	if((rv = _Write( sock, data, size )) < 1 ) {
		/* Error in writing the data */
		inet_close( sock );
		return(NULL);
	}

	buf = http_get_data_sock( sock, 0 );
	inet_close(sock);

	if( (rv = parse_http_headers_mem ( buf )) < 0 ) {
		PKI_log(PKI_LOG_ERR, "Can not parse HTTP headers!\n");
		if( buf ) PKI_MEM_free ( buf );
		return ( NULL );
	}

	if(( ret = PKI_STACK_MEM_new()) == NULL ) {
		return (NULL);
	}

	if((mm = PKI_MEM_new_null()) == NULL ) {
		if( buf ) PKI_MEM_free ( buf );
		return ( NULL );
	}

	PKI_MEM_add( mm, &(buf->data[rv]), buf->size - rv );

	PKI_STACK_MEM_push( ret, mm );

	PKI_log_debug("WARNING, Function URL_HTTP_post_data() not really "
							"implemented (yet)!");

	return( ret );
}
