/* Socket Wrapping functions */
/* OpenCA libpki package
* (c) 2000-2007 by Massimiliano Pala and OpenCA Group
* All Rights Reserved
*
* ===================================================================
* Released under OpenCA LICENSE
*/

#include <libpki/pki.h>

#define	LISTENQ		30

extern int h_errno;

int _Socket (int family, int type, int protocol) {
	int n;
	if ( (n = socket(family,type,protocol)) < 0)
		PKI_log( PKI_LOG_ERR, "Can not Init Socket (%s)",
			hstrerror(h_errno));
	return n;
}

int _Listen (char *hostname, int port) {
	int fd, reuse_addr;
	struct sockaddr_in servaddr;
	struct hostent *ht = NULL;

	if((fd = _Socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		return(0);
	};

	reuse_addr = 1;
	if(setsockopt(fd, SOL_SOCKET,  SO_REUSEADDR,  &reuse_addr, 
				sizeof(reuse_addr)) == -1 ) {
		PKI_log( PKI_LOG_ERR, "Can not Set Socket Options [%s]",
			hstrerror(h_errno));
		return(0);
	}
	memset( &servaddr, '\x0', sizeof( servaddr ));

	servaddr.sin_family = AF_INET;
	if( (hostname == NULL ) || (hostname[0] == '*')) {
		servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		if((ht = gethostbyname( hostname )) == NULL ) {
			PKI_log( PKI_LOG_ERR, "Can not Get Hostname [%s]",
				hstrerror(h_errno));
			return(0);
		}
		servaddr.sin_addr = 
				(**(struct in_addr **)ht->h_addr_list);
	}

	PKI_log ( PKI_LOG_INFO, "Setting Local Address %s [%d]\n", 
	 		inet_ntoa(servaddr.sin_addr), port );

	if ( port > 0 ) {
		servaddr.sin_port = htons( port );
	} else {
		servaddr.sin_port = 0;
	}

	if( bind(fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) == -1 ) {
		PKI_log(PKI_LOG_ERR, "Error in Socket Binding [%s]",
			hstrerror( h_errno ));
		return(0);
	};

	if( listen( fd, LISTENQ ) == -1 ) {
		PKI_log(PKI_LOG_ERR, "Can not listen to fd (%s)",
			hstrerror( h_errno ));
		return(0);
	};

	return(fd);
}

int _Accept (int listen_sockfd, struct sockaddr * cliaddr,
					socklen_t *addrlenp) {

	int n;
	struct sockaddr addr;
	socklen_t len;

again:

	if( !cliaddr || !addrlenp ) {
		len=sizeof(addr);
		n = accept(listen_sockfd, &addr, &len);
	} else {
		n = accept(listen_sockfd, cliaddr, addrlenp);
	}
		
	if ( n < 0) {
		if (INTERRUPTED_BY_SIGNAL)
			goto again;

		PKI_log(PKI_LOG_ERR,"Error while (ACCEPT) [%s]",
			hstrerror( h_errno ));
  	}
	return(n);
}

ssize_t _Read (int fd, void *bufptr, size_t nbytes) {

	ssize_t n;

again:
	if ( (n = read(fd,bufptr,nbytes)) < 0) {
		if (INTERRUPTED_BY_SIGNAL)
			goto again;
		else
			PKI_log(PKI_LOG_ERR, "Socket Read failed!");
	}
	return(n);
}

ssize_t _Write (int fd, void *bufptr, size_t nbytes) {

	ssize_t n;

again:
	if ( (n = write(fd,bufptr,nbytes)) < 0) {
		if (INTERRUPTED_BY_SIGNAL)
			goto again;
		else
			PKI_log(PKI_LOG_ERR, "Socket Write failed!");
	}
	return(n);
}

int _Select (int maxfdp1, fd_set *readset, fd_set *writeset, 
			fd_set *exceptset, struct timeval *timeout) {

	int n;
again:
	if ( (n = select (maxfdp1, readset, writeset, 
				exceptset, timeout)) < 0) {
		if (INTERRUPTED_BY_SIGNAL)
			goto again;
		else
			PKI_log( PKI_LOG_ERR, "Select failed!");
  	}
	return(n);
}

int _Connect (int sockfd, const SA *srvaddr, 
					socklen_t addrlen) {
	if ( connect(sockfd, srvaddr, addrlen) != 0) {
		PKI_log(PKI_LOG_ERR, "Socket Connect failed (%s)!", 
			hstrerror(h_errno));
		return(PKI_ERR);
	}
	return ( PKI_OK );
}


int _Close (int fd) {
  	if (close(fd) != 0) {
		PKI_log(PKI_LOG_ERR, "Socket Close failed (%s)!", 
			hstrerror(h_errno));
		return(0);
	}
	return( 1 );
}


void _Shutdown (int fd, int howto) {
	if (shutdown(fd,howto) != 0)
		PKI_log(PKI_LOG_ERR, "Socket Shutdown failed (%s)!", 
			hstrerror(h_errno));

	return;
}

struct hostent *Gethostbyname (const char *hostname) {

	struct hostent *hp = NULL;

	if( !hostname ) return NULL;
	if ((hp = gethostbyname(hostname)) == NULL)
		PKI_log( PKI_LOG_ERR, "Socket gethostbyname() failed for "
			"[%s]: %s", hostname, hstrerror(h_errno));
	return hp;
}

int inet_connect ( URL *url ) {

	int sockfd;
	int ret = 0;
	struct sockaddr_in *servaddr;
	struct hostent *hptr;

	/* create socket */
	if((sockfd = _Socket(AF_INET, SOCK_STREAM, 0)) == -1 ) {
		PKI_log( PKI_LOG_ERR, "Socket command failed (%s)",
			hstrerror(h_errno));
		return (-1);
	}

	/* specify address of server to connect to */
	if(( servaddr = (struct sockaddr_in *) 
			   malloc ( sizeof(struct sockaddr_in))) == NULL ) {
		PKI_log( PKI_LOG_ERR, "Memory allocation failed (%s:%d)",
						__FILE__, __LINE__ );
		return(-1);
	}

	if ((hptr = gethostbyname(url->addr)) == NULL) {
    		PKI_log( PKI_LOG_ERR, "gethostbyname() failed for [%s] : %s",
      			url->addr, hstrerror(h_errno));
		if( servaddr ) free (servaddr);
		return ( -1 );
	}

	memset (servaddr, 0, sizeof(struct sockaddr_in));
	servaddr->sin_family = hptr->h_addrtype;
	servaddr->sin_port   = htons(url->port);

	if (inet_aton(inet_ntoa(**((struct in_addr**)hptr->h_addr_list)), 
					&servaddr->sin_addr) == 0) {
		PKI_log( PKI_LOG_ERR, "Can not get server address (%s)",
			hstrerror(h_errno));
		if( servaddr ) free (servaddr);
		return( -1 );
	}

	/* try to connect */
	if(( ret = _Connect (sockfd, (struct sockaddr *) servaddr, 
				sizeof(struct sockaddr ))) == PKI_ERR ) {
		PKI_log( PKI_LOG_ERR, "Socket _Connect failed (%s)",
							hstrerror( h_errno ));

		if( servaddr ) free (servaddr);
		return(-1);
	}

	if( servaddr ) free (servaddr);

	PKI_log_debug( "Connection Successful to %s:%d", 
							url->addr, url->port );

	return ( sockfd );
}

int inet_close ( int fd ) {

	return _Close( fd );
}
