/* PKI_X509 object management */

#include <libpki/pki.h>
#include <sys/utsname.h>

PKI_X509_CERT * PKI_X509_CERT_new ( PKI_X509_CERT *ca_cert, PKI_KEYPAIR *pkey, 
			PKI_X509_REQ *req, char *subj_s, char *serial_s, 
				PKI_X509_PROFILE *conf, PKI_ALGOR *algor, 
					PKI_CONFIG *oids, HSM *hsm ) {

	PKI_X509_CERT *ret = NULL;
	PKI_X509_NAME *subj = NULL;
	PKI_X509_NAME *issuer = NULL;
	PKI_DIGEST_ALG *digest = NULL;

	int rv = 0;
	ASN1_INTEGER *serial = NULL;

	/* Check if the REQUIRED PKEY has been passed */
	if (!pkey) return (NULL);

	/* This has to be fixed, to work on every option */
	if( subj_s ) {
		subj = PKI_X509_NAME_new ( subj_s );
	} else if ( req ) {
		/* Copy the name from the request */
		subj = PKI_X509_REQ_get_attribute( req, PKI_ATTR_SUBJECT );
		// subj = PKI_X509_NAME_new( "CN=Shuld be copied from REQ, "
		// 						"O=OpenCA" );
	} else if ( conf ) {
		subj = PKI_X509_NAME_new( "CN=Shuld be copied from CONF, "
								"O=OpenCA" );
	} else {
		struct utsname myself;
                char tmp_name[1024];

                if (uname(&myself) < 0) {
			subj = PKI_X509_NAME_new( "O=OpenCA" );
                } else {
			sprintf( tmp_name, "CN=%s", myself.nodename );
			subj = PKI_X509_NAME_new( tmp_name );
		}
	}

	if( !subj ) {
		PKI_log_debug("ERROR, can not generate subject name! (%s)\n",
							subj_s );
		goto err;
	}

	if( ca_cert ) {
		PKI_X509_NAME *ca_subject = NULL;

		/* Let's get the ca_cert subject and dup that data */
		ca_subject = (PKI_X509_NAME *) 
				X509_get_subject_name( (X509 *) ca_cert );
		if( ca_subject ) {
			issuer = (PKI_X509_NAME *) X509_NAME_dup ( 
						(X509_NAME *) ca_subject );
		} else {
			PKI_log_debug("ERROR, can not get CA subj name!");
			goto err;
		}

	} else {
		issuer = (PKI_X509_NAME *) X509_NAME_dup((X509_NAME *) subj);
	}
	if( !issuer ) {
		PKI_log_debug("ERROR, no issuer name found!");
		goto err;
	}

	/* Alloc memory structure for the Certificate */
	ret = (PKI_X509_CERT *) X509_new();
	if( !ret ) {
		PKI_log_debug("ERROR, memory allocation!");
		return (NULL);
	}

	/* Set certificate version --- for simplicity let's set it to
	   v3, now */
	if (!X509_set_version((X509 *)ret,2)) {
		PKI_log_debug("ERROR, can not set cert version field!");
		goto err;
	}

	if( serial_s ) {
		serial = s2i_ASN1_INTEGER( NULL, serial_s);
	} else {
		serial = s2i_ASN1_INTEGER( NULL, "0");
	}

	if(!X509_set_serialNumber( (X509 *) ret, serial )) {
		PKI_log_debug("ERROR, can not set serial number (%s)!",
				serial_s );
		goto err;
	}

	/* Set the issuer Name */
	// rv = X509_set_issuer_name((X509 *) ret, (X509_NAME *) issuer);
	if(!X509_set_issuer_name((X509 *) ret, (X509_NAME *) issuer)) {
		PKI_log_debug("ERROR, can not set issuer name field!");
		goto err;
	}

	/* Set the subject Name */
	if(!X509_set_subject_name((X509 *) ret, (X509_NAME *) subj)) {
		PKI_log_debug("ERROR, can not set subject name field!");
		goto err;
	}

	/* Set the start date NOW */
	if(X509_gmtime_adj(X509_get_notBefore((X509 *) ret),0) == NULL ) {
		PKI_log_debug("ERROR, can not set notBefore field!");
		goto err;
	}
        // else ASN1_UTCTIME_set_string(X509_get_notBefore(ret),startdate);

	/* Set the end date in a year */
        if(X509_gmtime_adj(X509_get_notAfter((X509 *) ret),
					(long)60*60*24*365) == NULL ) {
		PKI_log_debug("ERROR, can not set notAfter field!");
		goto err;
	}

	/* Copy the PKEY if it is in the request, otherwise use the
	   public part of the PKI_X509_CERT */
	if( req ) {
		PKI_KEYPAIR *rkey = NULL;

		rkey = (PKI_KEYPAIR *) X509_REQ_get_pubkey((X509_REQ *) req);
		if( !rkey ) {
			PKI_log_debug("ERROR, can not get pubkey from req!");
			goto err;
		}

        	rv = X509_set_pubkey((X509 *) ret, (EVP_PKEY *) rkey);
		if( !rv ) {
			PKI_log_debug("ERROR, can not set pubkey in cert!");
			goto err;
		}

        	PKI_KEYPAIR_free(rkey, hsm);
		
	} else {
		/* Self Signed -- Same Public Key! */
		rv = X509_set_pubkey((X509 *) ret, (EVP_PKEY *) pkey);
		if( !rv ) {
			PKI_log_debug("ERROR, can not set key in self-signed"
				" cert!");
			goto err;
		}
	}

	if( conf ) {
		PKI_log_debug("INFO, Adding certificate extensions!");
		rv = PKI_X509_EXTENSIONS_cert_add_profile( conf, oids, 
								req, ret );
		if( !rv ) {
			PKI_log_debug( "ERROR, can not set extensions!");
			goto err;
		}
	}

	/* Let's set the digest for the right signature scheme */
	/*
        pp = (EVP_PKEY *) pkey;
	*/

	/*
	if( !algor ) {
        	switch( EVP_PKEY_type(pp->type) ) {
        	        case EVP_PKEY_DSA: 
	                        digest=PKI_DIGEST_ALG_DSS1;
				break;
#ifdef ENABLE_ECDSA
	                case EVP_PKEY_EC:
	                        digest=PKI_DIGEST_ALG_ECDSA_SHA1;
	                        break;
#endif
			default:
				digest = PKI_DIGEST_ALG_DEFAULT;
		}
	*/
	if( !algor ) {
		digest = PKI_DIGEST_ALG_get_by_key( pkey );
	} else {
		PKI_log_debug( "PKI_X509_CERT_new()::getting digest algor");
		digest = PKI_ALGOR_get_digest ( algor );
		PKI_log_debug( "PKI_X509_CERT_new()::got digest algor %p", digest);
	}

	if( digest == NULL ) {
		PKI_log_debug("%s:%d::PKI_X509_CERT_new()::ERROR, can "
			"not get the digest!", __FILE__, __LINE__ );

		return( NULL );
	};

	if(!PKI_X509_CERT_sign( ret, pkey, digest, hsm)) {
		PKI_log_debug("PKI_X509_CERT_new()::ERROR, can not sign certificate!");
		goto err;
	}

	return(ret);

err:
	PKI_log_debug("ERROR detected while generating a new cert!");

	if( ret ) PKI_X509_CERT_free ( ret );
	if( subj ) PKI_X509_NAME_free (subj);
	if( issuer ) PKI_X509_NAME_free (issuer);

	return (NULL);
}

void PKI_X509_CERT_free_void( void *x ) {
	PKI_X509_CERT_free((PKI_X509_CERT *)x);
}

int PKI_X509_CERT_free( PKI_X509_CERT *x ) {

	if( !x ) return (PKI_ERR);

	X509_free( (X509 *) x );

	return(PKI_OK);
}

int PKI_X509_CERT_add_extension(PKI_X509_CERT *x, PKI_X509_EXTENSION *ext) {

	if( !x || !ext ) return (PKI_ERR);

	if (!X509_add_ext((X509 *)x, ext, -1)) return (PKI_ERR);

	return (PKI_OK);
}

void * PKI_X509_CERT_get_data ( PKI_X509_CERT *x, PKI_X509_DATA type ) {

	void *ret = NULL;

	if( !x ) return (NULL);

	switch( type ) {
        	case PKI_X509_SERIAL:
			ret = X509_get_serialNumber ( (X509 *) x );
			break;
		case PKI_X509_SUBJECT:
			ret = X509_get_subject_name( (X509 *) x );
			break;
        	case PKI_X509_ISSUER:
			ret = X509_get_issuer_name( (X509 *) x );
			break;
		case PKI_X509_NOTBEFORE:
			ret = X509_get_notBefore( (X509 *) x );
			break;
		case PKI_X509_NOTAFTER:
			ret = X509_get_notAfter( (X509 *) x );
			break;
		case PKI_X509_KEYPAIR:
			ret = X509_get_pubkey( (X509 *) x);
			break;
        	case PKI_X509_ALGORITHM:
        	case PKI_X509_KEYSIZE:
        	case PKI_X509_BASIC_CONSTRAINS:
        	case PKI_X509_EXTENSIONS:
		default:
			/* Not Recognized/Supported DATATYPE */
			return (NULL);
	}

	return (ret);
}

int PKI_X509_CERT_print_parsed( PKI_X509_CERT *x, 
					PKI_X509_DATA type, int fd ) {

	char * data = NULL;
	int ret = 0;

	if((data = PKI_X509_CERT_get_parsed( x, type )) == NULL ) {
		return (PKI_ERR);
	} else {
		if( fd == 0 ) fd = 2;
		if(write(fd, data, strlen( data )) == -1 ) {
			ret = PKI_ERR;
		} else {
			ret = PKI_OK;
		}
		PKI_Free( data );
	}

	return (ret);
}

char * PKI_X509_CERT_get_parsed( PKI_X509_CERT *x, PKI_X509_DATA type ) {

	char *ret = NULL;

	if( !x ) return (NULL);

	switch( type ) {
		case PKI_X509_SUBJECT:
			ret = PKI_X509_NAME_get_parsed( (PKI_X509_NAME *) 
					X509_get_subject_name( (X509 *) x));
			break;
        	case PKI_X509_ISSUER:
			ret = PKI_X509_NAME_get_parsed( (PKI_X509_NAME *) 
					X509_get_issuer_name( (X509 *) x));
			break;
        	case PKI_X509_SERIAL:
			ret = PKI_INTEGER_get_parsed( (PKI_INTEGER *)
					X509_get_serialNumber( (X509 *) x));
			break;
		case PKI_X509_NOTBEFORE:
			ret = PKI_TIME_get_parsed( (PKI_TIME *)
				X509_get_notBefore((X509 *)x));
			break;
		case PKI_X509_NOTAFTER:
			ret = PKI_TIME_get_parsed( (PKI_TIME *)
				X509_get_notAfter((X509 *)x));
			break;
        	case PKI_X509_ALGORITHM:
        	case PKI_X509_KEYSIZE:
        	case PKI_X509_BASIC_CONSTRAINS:
        	case PKI_X509_EXTENSIONS:
		default:
			/* Not Recognized/Supported DATATYPE */
			return (NULL);
	}

	return (ret);
}

/*! \brief Calculates the fingerprint over a certificate by using the
 *         passed digest algorithm identifier
 */

PKI_DIGEST *PKI_X509_CERT_fingerprint ( PKI_X509_CERT *x, PKI_DIGEST_ALG *alg ){

	PKI_DIGEST *ret = NULL;
	unsigned char buf[EVP_MAX_MD_SIZE];
	int ret_size = 0;

	/* Check that we have a valid certificate */
	if( !x ) return ( NULL );

	/* If no Algorithm is provided, we use the default one */
	if( alg == NULL ) {
		alg = PKI_DIGEST_ALG_DEFAULT;
	}

	/* Calculate the Digest */
	if (!X509_digest(x,alg,buf,&ret_size)) {
		/* ERROR */
		return ( NULL );
	}

	/* Allocate the return structure */
        if((ret = PKI_Malloc ( sizeof( PKI_DIGEST) )) == NULL ) {
                /* Memory Allocation Error! */
                return( NULL );
        }

        /* Allocate the buffer */
        if((ret->digest = PKI_Malloc ( ret_size )) == NULL ) {
                /* Memory Error */
                PKI_Free ( ret );
                return( NULL );
        }

        /* Set the size of the Digest */
        ret->size = ret_size;

        /* Copy the Digest Data */
        memcpy( ret->digest, buf, ret->size );

	/* Sets the algorithm used */
	ret->algor = alg;

	return ( ret );

}

/*! \brief Calculates the fingerprint over a certificate by using the
 *         passed digest string (char *) identifier
 */

PKI_DIGEST *PKI_X509_CERT_fingerprint_by_name( PKI_X509_CERT *x, char *alg ) {

	PKI_DIGEST_ALG *alg_id = NULL;

	alg_id = PKI_DIGEST_ALG_get_by_name ( alg );

	return ( PKI_X509_CERT_fingerprint( x, alg_id ));
}
