/* PKI_X509 object management */

#include <libpki/pki.h>
#include <sys/utsname.h>

/*! \brief Generate a new CRL from a stack of revoked entries using a provided
 *  token
 *
 * Generates a new signed CRL from a stack of revoked entries. If a profile
 * passed, it is used to set the right extensions in the CRL. To generate a
 * new revoked entry the PKI_X509_CRL_ENTRY_new() function has to be used.
 */

PKI_X509_CRL *PKI_X509_CRL_new_tk ( PKI_TOKEN *tk, PKI_X509_CRL_ENTRY_STACK *sk,
		PKI_X509_PROFILE *profile ) {

	PKI_X509_CRL *crl = NULL;

	crl = PKI_X509_CRL_new( tk->keypair, tk->cert, sk, profile, tk->hsm );

	return( crl );
}

/*! \brief Generate a new CRL from a stack of revoked entries
 *
 * Generates a new signed CRL from a stack of revoked entries. A profile is
 * used to set the right extensions in the CRL. To generate a new revoked
 * entry the PKI_X509_CRL_ENTRY_new() function has to be used.
 */

PKI_X509_CRL *PKI_X509_CRL_new ( PKI_KEYPAIR *pkey, PKI_X509_CERT *cert, 
		PKI_X509_CRL_ENTRY_STACK *sk, PKI_X509_PROFILE *profile, 
								HSM *hsm ) {

	PKI_X509_CRL *crl = NULL;

	return( crl );
}

/*! \brief Generates a new PKI_X509_CRL_ENTRY from a certificate
 *
 * This function generates a new PKI_X509_CRL_ENTRY starting from a certificate.
 * The new structure can be added to a PKI_X509_CRL_ENTRY_STACK structure in
 * order to generate a new CRL.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_ENTRY_new ( PKI_X509_CERT *cert, 
					PKI_X509_PROFILE *profile ) {

	PKI_X509_CRL_ENTRY *entry = NULL;

	if((entry = (PKI_X509_CRL_ENTRY *) X509_REVOKED_new()) == NULL ) {
		/* Memory alloc error */
		return (NULL);
	}

	if((entry->serialNumber = 
		ASN1_INTEGER_dup( X509_get_serialNumber((X509 *)cert)))
								== NULL ) {
		goto err;
	}

	return( entry );

err:
	if( entry ) X509_REVOKED_free ( (X509_REVOKED *) entry );
	return (NULL);
}

/*! \brief Generates a new PKI_X509_CRL_ENTRY from a serial number (string)
 *
 * This function generates a new PKI_X509_CRL_ENTRY starting from a string
 * representing the serial number of the entry. The optional profile
 * is used to add extensions to the entry (when needed)
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_ENTRY_new_serial ( char *serial, 
					PKI_X509_PROFILE *profile ) {

	PKI_X509_CRL_ENTRY *entry = NULL;

	if( !serial ) return (NULL);

	if((entry = (PKI_X509_CRL_ENTRY *) X509_REVOKED_new()) == NULL ) {
		/* Memory alloc error */
		return (NULL);
	}

	if((entry->serialNumber = s2i_ASN1_INTEGER( NULL, serial )) == NULL ) {
		goto err;
	}
	return( entry );

err:

	if( entry ) X509_REVOKED_free ( (X509_REVOKED *) entry );
	return (NULL);
}

void PKI_X509_CRL_ENTRY_free_void ( void *entry ) {
	PKI_X509_CRL_ENTRY_free ( (PKI_X509_CRL_ENTRY *) entry );
}

/*! \brief Frees a PKI_X509_CRL_ENTRY
 *
 * Frees memory associated to a PKI_X509_CRL_ENTRY
 */

int PKI_X509_CRL_ENTRY_free ( PKI_X509_CRL_ENTRY *entry ) {

	if( !entry ) return (PKI_ERR);

	if( entry ) X509_REVOKED_free ( (X509_REVOKED *) entry );

	return (PKI_OK);

}

/*! \brief Find an entry within a CRL
 *
 * This function look inside a CRL and returns the PKI_X509_CRL_ENTRY
 * if the passed serial is found.  The returned pointer refers to the
 * internal CRL structure, when the CRL is freed the pointer will no
 * more point to a valid memory area.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_lookup_serial( PKI_X509_CRL *x, char *serial ) {

	ASN1_INTEGER *s = NULL;
	PKI_X509_CRL_ENTRY *r = NULL;
	// PKI_X509_CRL_ENTRY *ret = NULL;
	X509_CRL *crl = NULL;

        unsigned long curr = 0;
        // unsigned long start = 0;
        // int cont = 1;
        unsigned long end, cmp_val;
        // int found = 0;

	if( !x || !serial ) return (NULL);

	crl = (X509_CRL *) x;

	/* Set the end point to the last one */
        end = sk_X509_REVOKED_num(crl->crl->revoked) - 1;
        if( end < 0 ) return (r);

	s = ASN1_INTEGER_new();
	if( !s ) return (0);

	if((s = s2i_ASN1_INTEGER( NULL, serial )) == NULL ) {
		ASN1_INTEGER_free ( s );
		return(0);
	}

	for( curr = 0 ; curr <= end ; curr++ ) {
		r = sk_X509_REVOKED_value( crl->crl->revoked, curr );
                if(( cmp_val = ASN1_INTEGER_cmp(r->serialNumber, s)) == 0 ) {
			/* Found */
			break;
		}
	}

	if( s ) ASN1_INTEGER_free (s);

	// ret = X509_REVOKED_dup( (X509_REVOKED *) r );

	return( r );
}

/*! \brief Find an entry in a CRL by using a certificate
 *
 * Use the information within the certificate to look for a corresponding
 * entry in the CRL. If found, the PKI_X509_CRL_ENTRY structure is copied and
 * returned.  The returned pointer refers to the internal CRL structure, 
 * when the CRL is freed the pointer will no more point to a valid memory
 * area.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_lookup_cert ( PKI_X509_CRL *x, PKI_X509_CERT *cert ) {

	ASN1_INTEGER *serial = NULL;
	PKI_X509_CRL_ENTRY *entry = NULL;
	char *serial_s = NULL;

	if( !x || !cert ) return (NULL);

	if((serial = X509_get_serialNumber( (X509 *) cert )) == NULL ) {
		return (NULL);
	}

	if((serial_s = i2s_ASN1_INTEGER( NULL, serial )) == NULL ) {
		return(NULL);
	}

	entry = PKI_X509_CRL_lookup_serial( x, serial_s );

	/* Now let's free the serial_s */
	if( serial_s ) PKI_Free ( serial_s );

	return( entry );
}

void PKI_X509_CRL_free_void ( void * x ) {
	PKI_X509_CRL_free ( (PKI_X509_CRL *) x );
}

/*! \brief Frees memory associated with a PKI_CRL
 *
 * This function frees memory associated with a PKI_X509_CRL data structure.
 * Returns PKI_OK if successful or PKI_ERR in case of an error (or if
 * the passed pointer was NULL.
 */

int PKI_X509_CRL_free ( PKI_X509_CRL * x ) {

	if( !x ) return (PKI_ERR);

	if( x ) X509_CRL_free ((X509_CRL *) x );

	return( PKI_OK );
}

int PKI_X509_CRL_add_extension(PKI_X509_CRL *x, PKI_X509_EXTENSION *ext) {

	if( !x || !ext ) return (PKI_ERR);

	if (!X509_CRL_add_ext((X509_CRL *)x, ext, -1)) return (PKI_ERR);

	return (PKI_OK);
}

void * PKI_X509_CRL_get_data ( PKI_X509_CRL *x, PKI_X509_DATA type ) {

	void *ret = NULL;

	if( !x ) return (NULL);

	switch( type ) {
		case PKI_X509_SUBJECT:
        	case PKI_X509_ISSUER:
        	case PKI_X509_SERIAL:
        	case PKI_X509_ALGORITHM:
        	case PKI_X509_KEYSIZE:
        	case PKI_X509_BASIC_CONSTRAINS:
        	case PKI_X509_EXTENSIONS:
		default:
			/* Not Recognized/Supported DATATYPE */
			return (NULL);
	}

	return (ret);
}

const char * PKI_X509_CRL_get_parsed( PKI_X509_CRL *x, PKI_X509_DATA type ) {

	const char *ret = NULL;

	if( !x ) return (NULL);

	switch( type ) {
		case PKI_X509_SUBJECT:
			return "";
			break;
        	case PKI_X509_ISSUER:
			ret = PKI_X509_NAME_get_parsed( (PKI_X509_NAME *) 
				X509_CRL_get_issuer( (X509_CRL *) x));
			break;
        	case PKI_X509_SERIAL:
        	case PKI_X509_ALGORITHM:
        	case PKI_X509_KEYSIZE:
        	case PKI_X509_BASIC_CONSTRAINS:
        	case PKI_X509_EXTENSIONS:
		default:
			/* Not Recognized/Supported DATATYPE */
			return (NULL);
	}

	return (ret);
}
