/* PKI_X509 object management */

#include <libpki/pki.h>
#include <sys/utsname.h>

PKI_X509_REQ *PKI_X509_REQ_new ( PKI_KEYPAIR *pkey, char *subj_s,
			PKI_X509_PROFILE *req_cnf, PKI_CONFIG *oids, 
				PKI_DIGEST_ALG *digest, HSM *hsm) {

	PKI_X509_REQ *req = NULL;
	EVP_PKEY *pp = NULL;
	int rv = PKI_OK;

	PKI_X509_NAME *name = NULL;

	/* We need at least the private key for the request */
	if( !pkey ) {
		PKI_log_debug("ERROR, no key for PKI_X509_REQ_new()!");
		return (NULL);
	}
	pp = (EVP_PKEY *) pkey;

	/* Let's set the digest for the right signature scheme */
	if( !digest ) {
		digest = PKI_DIGEST_ALG_get_by_key( pkey );
	};

	if( digest == NULL ) {
		PKI_log_debug("%s:%d::PKI_X509_REQ_new()::ERROR, can "
			"not get the default digest!", __FILE__, __LINE__ );

		return( NULL );
	};

	/*
		switch( EVP_PKEY_type(pp->type) ) {
			case EVP_PKEY_DSA:
				digest = PKI_DIGEST_ALG_DSS1;
				PKI_log_debug("[DSA] DIGEST => %p", digest );
				break;
#ifdef ENABLE_ECDSA
			case EVP_PKEY_EC:
				digest=PKI_DIGEST_ALG_ECDSA_DSS1;
				PKI_log_debug("[EC] DIGEST => %p", digest );
				break;
#endif
			default:
				digest=PKI_DIGEST_ALG_DEFAULT;
				PKI_log_debug("[DEF] DIGEST => %p", digest );
		}
	};
	*/


	/* Generate the request */
	if((req = (PKI_X509_REQ *) X509_REQ_new()) == NULL ) {
		PKI_log_debug("Memory Allocation error [%s:%d]",
						__FILE__, __LINE__ );
		return ( NULL );
	}

	if( subj_s ) {
		name = PKI_X509_NAME_new ( subj_s );
	} else if( req_cnf ) {
		/* TODO: Process the data by adding fields, exts, etc.. to
		   the request */
		name = PKI_X509_NAME_new ( "CN=Should be taken from config, "
						"O=OpenCA" );
	} else {
		struct utsname myself;
                char tmp_name[BUFF_MAX_SIZE];

                if (uname(&myself) < 0) {
			name = PKI_X509_NAME_new( "O=OpenCA" );
                } else {
			snprintf( tmp_name, BUFF_MAX_SIZE - 1, 
				"CN=%s, O=OpenCA", myself.nodename );
			name = PKI_X509_NAME_new( tmp_name );
		}
	}

	if( !name ) {
		PKI_log_debug("No Name generated for new request!");
	} else {

		if (!X509_REQ_set_subject_name( (X509_REQ *) req, 
							(X509_NAME *) name)) {
			if( name ) X509_NAME_free ( (X509_NAME *) name);
			PKI_log_debug("Can not set Name for new request!");
			goto err;
		}
	}

	/* Set the version number and the pubkey */
	if (!X509_REQ_set_version( (X509_REQ *) req, 2L) ||
			!X509_REQ_set_pubkey( (X509_REQ *) req, 
							(EVP_PKEY *) pkey) ) {
		if( req ) X509_REQ_free( (X509_REQ *) req );
		PKI_log_debug("ERROR, X509_REQ Can not set ver or pubkey");
		return (NULL);
	}

	if( req_cnf != NULL ) {
		/* Add Request Extensions */
		rv = PKI_X509_EXTENSIONS_req_add_profile( req_cnf, oids, req );
	}

	rv = PKI_X509_REQ_sign( req, pkey, digest, hsm );
	if (rv == PKI_ERR ) {
		/* Error Signing the request */
		PKI_log_debug("REQ::ERROR signing the Request -- mod");
		ERR_print_errors_fp( stderr );
		if( req ) X509_REQ_free( (X509_REQ *) req );
		return (NULL);
	}

	/* Return the requested PKI_X509_REQ structure */
	return req;

err:
	if( req ) PKI_X509_REQ_free ( req );
	return (NULL);

}

void PKI_X509_REQ_free_void( void *x ) {
	PKI_X509_REQ_free( (PKI_X509_REQ *)x);
}

int PKI_X509_REQ_free( PKI_X509_REQ *x ) {

	/* if no PKI_X509_REQ is passed, let's return 0 */
	if( !x ) return (PKI_ERR);

	/* Free the memory */
	X509_REQ_free( (X509_REQ *) x );

	/* Return success -- 1 */
	return(PKI_OK);
}

int PKI_X509_REQ_add_ext_stack(PKI_X509_REQ *x, PKI_X509_EXTENSION_STACK *ext) {

	STACK_OF(X509_REQ) *sk = NULL;
	int i = 0;

	if( !x || !ext ) return (PKI_ERR);

	if((sk = sk_X509_EXTENSION_new_null()) == NULL ) return (PKI_ERR);

	for( i = 0; i < PKI_STACK_X509_EXTENSION_elements(ext); i++ ) {
		sk_X509_EXTENSION_push( sk, 
			PKI_STACK_X509_EXTENSION_get_num( ext, i) );
	}

	if( !X509_REQ_add_extensions((X509_REQ *)x, sk )) return (PKI_ERR);

	if( sk ) sk_X509_EXTENSION_pop_free ( sk, X509_EXTENSION_free );

	return (PKI_OK);
}


void * PKI_X509_REQ_get_attribute ( PKI_X509_REQ *req, 
					PKI_X509_ATTRIBUTE attr ) {
	
	void *ret = NULL;

	if( !req ) return (NULL);

	switch( attr ) {
		case PKI_ATTR_SUBJECT:
			ret = (void *) X509_REQ_get_subject_name ( 
							(X509_REQ *) req );
			break;
		default:
			return( NULL );
	}

	return (ret);
}

