/* Config management for libpki */

#include <libpki/pki.h>

#include <sys/types.h>
#include <dirent.h>

/* Static function, to be used only internally */
static char * _xml_search_namespace_add ( char *search );

void logXmlMessages( void *userData, xmlErrorPtr error ) {
	PKI_log_debug( "XML I/O Error: %s", error->message);
	return;
}

PKI_CONFIG * PKI_CONFIG_load(char *urlPath) {

    PKI_CONFIG *doc = NULL;
    URL *url = NULL;
    xmlParserCtxt *parserCtxt = NULL;

    LIBXML_TEST_VERSION

    if( urlPath ) {
        url = URL_new( urlPath );
    } else {
	return ( NULL );
    }

    if( !url ) {
	return(PKI_ERR);
    }

    if((parserCtxt = xmlNewParserCtxt()) == NULL ) {
	URL_free( url );
	return(PKI_ERR);
    }

    xmlSetStructuredErrorFunc( parserCtxt, logXmlMessages );

    /* Do not Keep Blank Nodes */
    xmlKeepBlanksDefault(0);

    /*parse the file and get the DOM */
    doc = (PKI_CONFIG *) xmlCtxtReadFile(parserCtxt, url->addr, NULL, 
			XML_PARSE_RECOVER | XML_PARSE_NOERROR | 
				XML_PARSE_NOWARNING | XML_PARSE_NOENT );

    // xmlSkipBlankChars( parserCtxt );

    /*
    if( !doc ) {
	fprintf(stderr, "DEBUG:: NO DOC!\n");
    }
    */

    return( doc );

}

void PKI_CONFIG_free_void ( void * doc ) {
	PKI_CONFIG *my_doc = NULL;
	int i = 0;

	my_doc = (PKI_CONFIG *) doc;

	i = PKI_CONFIG_free( my_doc);

	return;
}

int PKI_CONFIG_free ( PKI_CONFIG * doc ) {
	if( !doc ) return (PKI_OK);

	xmlFreeDoc( doc );

	/*
	*Free the global variables that may
	*have been allocated by the parser.
	*/
	//xmlCleanupParser();

	return( PKI_OK );
}

PKI_OID * PKI_CONFIG_OID_search ( PKI_CONFIG *doc, char *searchName ) {

	PKI_OID *oid = NULL;
	PKI_CONFIG_ELEMENT *curr = NULL;

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	xmlChar oidSearchBuff[BUFF_MAX_SIZE];

	int size = 0;
	int i = 0;

	if( !doc || !searchName ) return (NULL);

	if((oid = PKI_OID_get( searchName )) != NULL ) {
		return ( oid );
	}

	snprintf((char *) oidSearchBuff, BUFF_MAX_SIZE, 
		"/%s:objectIdentifiers/%s:oid[@name=\"%s\"]",
			PKI_NAMESPACE_PREFIX, PKI_NAMESPACE_PREFIX, 
				searchName);

	xpathCtx = xmlXPathNewContext(doc);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(NULL);
	}

	xmlXPathRegisterNs(xpathCtx, 
		(xmlChar *) PKI_NAMESPACE_PREFIX, 
		(xmlChar *) PKI_NAMESPACE_HREF);

	xpathObj = xmlXPathEvalExpression( (xmlChar *) oidSearchBuff, xpathCtx);
	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		return(NULL);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	if( size > 1 ) {
		PKI_log_debug("WARNING, multiple %s found (OID searching) - "
				"using first!\n", searchName );
	}

	for( i = 0; i < size; i++ ) {
		if( nodes->nodeTab[i]->type == XML_ELEMENT_NODE ) {
			xmlChar *name = NULL;
			xmlChar *descr = NULL;
			xmlChar *val = NULL;

			curr = nodes->nodeTab[i];

			name = xmlGetProp( curr, (xmlChar *) "name" );
			descr = xmlGetProp( curr, (xmlChar *) "description" );
			val = xmlNodeListGetString(doc, 
						curr->xmlChildrenNode, 1);

			oid = PKI_OID_new ( (char *) val, (char *) name, 
							(char *) descr);

			if( descr ) xmlFree ( descr  );
			if( name ) xmlFree ( name );
			if( val ) xmlFree ( val );

			if( oid != NULL ) break;
		}
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	return (oid);
}

PKI_STACK * PKI_CONFIG_get_stack_value ( PKI_CONFIG *doc, char *search ) {

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	PKI_STACK *sk = NULL;

	char *val = NULL;
	int size = 0;
	int i = 0;

	char *my_search = NULL;

	if( !doc || !search ) return (NULL);

	xpathCtx = xmlXPathNewContext(doc);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(NULL);
	}

	my_search = _xml_search_namespace_add ( search );

	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
				(xmlChar *) PKI_NAMESPACE_HREF);

	xpathObj = xmlXPathEvalExpression( (xmlChar *) my_search, xpathCtx);
	PKI_Free ( my_search );

	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		return(NULL);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	sk = PKI_STACK_new(NULL);
	for( i=0; i < size; i++ ) {
		PKI_CONFIG_ELEMENT *curr = NULL;

		curr = nodes->nodeTab[i];
		val = (char *) xmlNodeListGetString(doc, 
						curr->xmlChildrenNode, 1);

		PKI_STACK_push( sk, strdup( val ));
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	return (sk);
}

static char * _xml_search_namespace_add ( char *search ) {

	char *my_search = NULL;
	char *my_arg = NULL;
	char *ret = NULL;

	int r = 0;
	int i = 0;

	/* Let's alloc enough memory for the arguments, maybe this is
	   too much, but for the moment, let's keep it big */
	my_arg = PKI_Malloc ( BUFF_MAX_SIZE );
	my_search = PKI_Malloc ( BUFF_MAX_SIZE );

	/* Now let's take care about setting the appropriate namespace
	   if it is not passed, already */
	i = 0;
	while( search[i] == '/' ) {
		i++;
		strncat(my_search, "/", BUFF_MAX_SIZE );
	}

	while( (i < strlen( search )) &&
			 (sscanf( search + i, "%[^/]%n", my_arg, &r ) > 0 )) {
		i = i + r;

		if( strchr( my_arg, ':' ) == NULL ) {
			strncat( my_search, PKI_NAMESPACE_PREFIX ":",
					BUFF_MAX_SIZE - strlen(my_search) );
		}
		strncat( my_search, my_arg, BUFF_MAX_SIZE - strlen(my_search));

		while( search[i] == '/' ) {
			i++;
			strncat(my_search, "/", 
				BUFF_MAX_SIZE - strlen( my_search ));
		}
	}
	PKI_Free( my_arg );

	ret = PKI_Malloc ( strlen( my_search ) + 1);
	strncpy( ret, my_search, strlen(my_search) );

	PKI_Free ( my_search );
	return( ret );
}

char * PKI_CONFIG_get_value ( PKI_CONFIG *doc, char *search ) {

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	char *val = NULL;
	int size = 0;
	int i = 0;

	char *my_search = NULL;

	if( !doc || !search ) return (NULL);

	xpathCtx = xmlXPathNewContext(doc);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(NULL);
	}

	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
				(xmlChar *) PKI_NAMESPACE_HREF);

	my_search = _xml_search_namespace_add ( search );

	xpathObj = xmlXPathEvalExpression( (xmlChar *) my_search, xpathCtx);
	PKI_Free (my_search);

	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		return(NULL);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	if( size >= 1 ) {
		PKI_CONFIG_ELEMENT *curr = NULL;

		curr = nodes->nodeTab[0];
		val = (char *) xmlNodeListGetString(doc, 
						curr->xmlChildrenNode, 1);
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	return (val);
}

PKI_CONFIG_ELEMENT * PKI_CONFIG_get_element ( PKI_CONFIG *doc, char *search ) {

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	PKI_CONFIG_ELEMENT *curr = NULL;

	int size = 0;
	int i = 0;

	char *my_search = NULL;

	if( !doc || !search ) return (NULL);

	xpathCtx = xmlXPathNewContext(doc);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(NULL);
	}

	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
					(xmlChar *) PKI_NAMESPACE_HREF);

	my_search = _xml_search_namespace_add ( search );

	xpathObj = xmlXPathEvalExpression( (xmlChar *) my_search, xpathCtx);
	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		PKI_Free ( my_search );
		return(NULL);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	if( size >= 1 ) {
		/* recursively copy the node */
		curr = xmlCopyNode( nodes->nodeTab[i], 1);
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	PKI_Free ( my_search );

	return (curr);
}

char * PKI_CONFIG_get_element_value (PKI_CONFIG_ELEMENT *e) {

	if( !e ) return (NULL);

	return( (char *) xmlNodeGetContent( (xmlNodePtr) e ) );
}

char * PKI_CONFIG_get_element_name (PKI_CONFIG_ELEMENT *e) {

	if( !e ) return (NULL);

	return( (char *) e->name );
}

PKI_CONFIG_ELEMENT * PKI_CONFIG_get_element_child ( PKI_CONFIG_ELEMENT *e) {

	PKI_CONFIG_ELEMENT *ret = NULL;

	if( !e || !e->children ) return (NULL);
	ret = e->children;

	while( ret && ret->type != XML_ELEMENT_NODE ) {
		ret = ret->next;
	}

	return( ret );
}

PKI_CONFIG_ELEMENT * PKI_CONFIG_get_element_next ( PKI_CONFIG_ELEMENT *e) {

	PKI_CONFIG_ELEMENT *ret = NULL;

	if( !e || !e->next ) return (NULL);

	ret = e->next;
	while( ret && ret->type != XML_ELEMENT_NODE ) {
		ret = ret->next;
	}

	return( ret );
}

PKI_CONFIG_ELEMENT * PKI_CONFIG_get_element_prev ( PKI_CONFIG_ELEMENT *e) {

	PKI_CONFIG_ELEMENT *ret = NULL;

	if( !e || !e->prev ) return (NULL);

	ret = e->prev;
	while( ret && ret->type != XML_ELEMENT_NODE ) {
		ret = ret->prev;
	}

	return( (PKI_CONFIG_ELEMENT *) e->prev );
}

PKI_CONFIG_ELEMENT_STACK * PKI_CONFIG_get_element_stack ( PKI_CONFIG *doc, 
							char *search ) {

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	PKI_CONFIG_ELEMENT *curr = NULL;
	PKI_CONFIG_ELEMENT_STACK *ret = NULL;

	int size = 0;
	int i = 0;

	char *my_search = NULL;

	if( !doc || !search ) return (NULL);

	xpathCtx = xmlXPathNewContext(doc);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(NULL);
	}

	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
					(xmlChar *) PKI_NAMESPACE_HREF);

	my_search = _xml_search_namespace_add ( search );

	xpathObj = xmlXPathEvalExpression( (xmlChar *) my_search, xpathCtx);
	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		PKI_Free ( my_search );
		return(NULL);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	ret = PKI_STACK_CONFIG_ELEMENT_new();
	if( size >= 1 ) {
		/* recursively copy the node */
		for( i = 0; i <size; i++ ) {
			if( nodes->nodeTab[i]->type != XML_ELEMENT_NODE )
				continue;
			curr = xmlCopyNode( nodes->nodeTab[i], 1);
			PKI_STACK_CONFIG_ELEMENT_push( ret, curr );
		}
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	PKI_Free ( my_search );

	return (ret);
}

PKI_CONFIG_ELEMENT_STACK * PKI_CONFIG_get_element_children ( 
					PKI_CONFIG_ELEMENT *e) {

	PKI_CONFIG_ELEMENT_STACK *ret = NULL;
	PKI_CONFIG_ELEMENT *curr = NULL;

	if( !e ) return (NULL);

	if( (curr = e->children) == NULL ) return (NULL);

	if((ret = PKI_STACK_CONFIG_ELEMENT_new()) == NULL ) {
		return( NULL );
	}

	while( curr ) {
		if( curr->type != XML_ELEMENT_NODE ) continue;

		PKI_STACK_CONFIG_ELEMENT_push( ret, curr );
		curr = curr->next;
	}

	return( ret );
}

/*!
 * \brief Returns a pointer to the filename of the configuration file that
          contains the configuration named 'name'.
 */

char * PKI_CONFIG_find ( char *dir, char *name ) {

	char buff[BUFF_MAX_SIZE];

        struct dirent *dd = NULL;
	DIR *dirp = NULL;
	URL *url = NULL;

	int found = 0;
	char *ret = NULL;

	/* Check input */
	if( !dir || !name ) {
		return (PKI_ERR);
	}

	if(( url = URL_new ( dir )) == NULL ) {
		return (PKI_ERR);
	}

	if( url->proto != URI_PROTO_FILE ) {
		return (PKI_ERR);
	}

	if((dirp = opendir( url->addr )) == NULL ) {
		return (PKI_ERR);
	} else {
		while(( dd = readdir( dirp )) != NULL ) {
			long len;
			char *filename = NULL;

			filename = dd->d_name;
			len = (long) strlen( filename );

			if( (len < 4) || (strcmp( ".xml", filename +len -4 ))) {
				continue;
			} else {
			
				char fullpath[BUFF_MAX_SIZE];
				long fullsize = 0;

				PKI_CONFIG *tmp_cfg = NULL;
				char *tmp_name = NULL;

				snprintf(fullpath, BUFF_MAX_SIZE,
					"%s/%s", url->addr, filename );

				if((fullsize = strlen(url->addr) + 
					strlen( filename ) + 1) > 
							BUFF_MAX_SIZE) {
					continue;
				}
				
				if((tmp_cfg = PKI_CONFIG_load( fullpath )) ==
									NULL ) {
					continue;
				}

				if((tmp_name = PKI_CONFIG_get_value( tmp_cfg,
						"/*/name")) != NULL) {
					if( strncmp_nocase(tmp_name, name,
						    strlen(tmp_name)) == 0) {
						ret = strdup( fullpath );
						found = 1;
						break;
					}
				}
			}
		}
		closedir( dirp );
	}
	if( url ) URL_free (url);

	if( found == 1 ) {
		return( ret );
	} else {
		return ( NULL );
	}
}

/*!
 * \brief Returns a stack of all configuration files found in the passed
 *        directory.
 */

PKI_CONFIG_STACK * PKI_CONFIG_load_dir ( char *dir ) {

	char buff[BUFF_MAX_SIZE];

        struct dirent *dd = NULL;
	DIR *dirp = NULL;
	URL *url = NULL;

	int found = 0;
	PKI_CONFIG_STACK *ret = NULL;

	/* Check input */
	if( !dir ) {
		return (NULL);
	}

	if(( url = URL_new ( dir )) == NULL ) {
		PKI_log_debug( "Dir not valid for config (%s)", dir );
		return ( NULL );
	}

	if( url->proto != URI_PROTO_FILE ) {
		PKI_log_debug( "Dir not valid for config (%s)", dir );
		return (NULL);
	}

	if((dirp = opendir( url->addr )) == NULL ) {
		PKI_log_debug("ERROR, Can not open dir %s!\n", url->addr );
		return (NULL);
	} else {
		if((ret = PKI_STACK_CONFIG_new()) == NULL ) {
			PKI_log_debug("Memory Error (%s:%d)", 
							__FILE__, __LINE__ );
			return(NULL);
		}
		while(( dd = readdir( dirp )) != NULL ) {
			long len;
			char *filename = NULL;

			filename = dd->d_name;
			len = (long) strlen( filename );

			if( (len < 4) || (strcmp( ".xml", filename +len -4 ))) {
				PKI_log_debug( "Skipping file %s", filename);
				continue;
			} else {
			
				char fullpath[BUFF_MAX_SIZE];
				long fullsize = 0;

				PKI_CONFIG *tmp_cfg = NULL;
				char *tmp_name = NULL;

				PKI_log_debug( "Loading file %s/%s", 
							url->addr, filename );

				snprintf(fullpath, BUFF_MAX_SIZE,
					"%s/%s", url->addr, filename );

				if((fullsize = strlen(url->addr) + 
					strlen( filename ) + 1) > 
							BUFF_MAX_SIZE) {
					continue;
				}
				
				if((tmp_cfg = PKI_CONFIG_load( fullpath )) ==
									NULL ) {
					continue;
				}

				PKI_log_debug( "Loaded %s file", fullpath );
				PKI_STACK_CONFIG_push( ret, tmp_cfg );
				found = 1;
			}
		}
		closedir( dirp );
	}
	if( url ) URL_free (url);

	if( found == 1 ) {
		return (ret);
	} else {
		PKI_STACK_CONFIG_free( ret );
		PKI_log_debug("PKI_CONFIG_load_dir() Failed!\n" );
		return ( NULL );
	}
}
