/* PKI Resource Query Protocol Message implementation
 * (c) 2006 by Massimiliano Pala and OpenCA Group
 * All Rights Reserved
 *
 * This software is released under the GPL2 License included
 * in the archive. You can not remove this copyright notice.
 */
                                                                                
#include <libpki/pki.h>

/*!
 * \brief Retireve a PKI_STACK of addresses from a PRQP server.
 *
 * Retrieve informations about services provided by a CA from a PRQP server.
 * The function returns a stack of strings containing URLs of the requested
 * services. If no URL (char *) is passed, then the library will search for
 * the default config file /etc/pki.conf for the configured Resource Query
 * Authority (PRQP Server).
 *
 */

PKI_STACK * PKI_get_ca_resources(PKI_X509_CERT *caCert, 
			PKI_X509_CERT *caIssuerCert, PKI_X509_CERT *issuedCert,
			PKI_STACK *sk_services, char *url_s ) {

	PRQP_REQ *p = NULL;
	PRQP_RESP *r = NULL;
	URL *url = NULL;
	PKI_STACK * url_sk = NULL;

	p = PRQP_REQ_new_certs_res ( caCert, caIssuerCert, issuedCert, 
								sk_services );
	if( !p ) return ( NULL );

	if( url_s ) {
		url = URL_new (url_s);
	}

	r = PKI_get_PRQP_RESP_url( p, url );

	if( url ) URL_free ( url );

	return PRQP_RESP_url_sk(r);
}

/*!
 * \brief Retireve the first configured URL for a PKI service from a PRQP
 *        server
 *
 * Retrieve informations about a specific service provided by a CA.
 * The function returns a string containing the URL of the requested
 * service (if available). 
 *
 * If no URL (char *) is passed, then the library will search for
 * the default config file /etc/pki.conf for the configured Resource Query
 * Authority (PRQP Server).
 *
 */

char * PKI_get_ca_service( PKI_X509_CERT *caCert, char *srv, char *url_s ) {

	PKI_STACK *services = NULL;
	PKI_STACK *ret_sk = NULL;

	char *ret_s = NULL;

	if( !srv || !caCert ) return ( NULL );

	if((services = PKI_STACK_new_null()) == NULL ) {
		PKI_log_debug("Stack creation error in %s:%d",
						__FILE__, __LINE__ );
		return ( NULL );
	}

	PKI_STACK_push( services, srv );
	ret_sk = PKI_get_ca_resources( caCert, NULL, NULL, services, url_s );

	PKI_STACK_free_all( services );

	if( !ret_sk ) {
		return ( NULL );
	} else {
		ret_s = PKI_STACK_pop( ret_sk );
		PKI_STACK_free_all( ret_sk );
	}

	return ( ret_s );
}

/*!
 * \brief Retireve a PRQP response from the passed url or from one of the
 *        configured RQAs in /etc/pki.conf
 */

PRQP_RESP * PKI_get_PRQP_RESP( PRQP_REQ *p, char *url_s ) {

	URL *url = NULL;

	if( p == NULL ) return (NULL);

	if( url_s != NULL ) {
		if((url = URL_new( url_s )) == NULL) {
			return(NULL);
		}
	}

	return( PKI_get_PRQP_RESP_url( p, url ));
}

/*!
 * \brief Retrieve a PRQP Response from a server.
 *
 * The function returns a PRQP_RESP if succesful or NULL if an error occurs
 * when contacting the PRQP server.
 */

PRQP_RESP * PKI_get_PRQP_RESP_url( PRQP_REQ *p, URL *url ) {

	PRQP_RESP * r = NULL;

	char line[1024], name[1024], addr[1024];

        FILE *file;

	if( !p ) {
		PKI_log_debug( "WARNING, no PRQP request when trying to get"
				" the response!");
		return ( NULL );
	}

	if( url ) {
		r = PRQP_http_get_resp ( url, p, 0);
		return( r );
	}

        file = fopen( PRQP_LIB_CONF_FILE, "r");
        if( !file ) {
		PKI_log_debug( "WARNING, PRQP config file %s not found!",
			PRQP_LIB_CONF_FILE );
		return ( NULL );
        }

       	while(!feof(file)) {
               	if( fgets(line, sizeof(line), file) ) {
                       	if((memcmp(line, ";", 1) == 0) || 
					(memcmp(line, "#", 1) == 0))
                                		continue;

			if(sscanf(line, "%s %s", name, addr ) > 1 ) {
				if((strcmp_nocase( name, 
					PRQP_LIB_CONF_ENTRY_LONG)==0) ||
						(strcmp_nocase ( name, 
						   PRQP_LIB_CONF_ENTRY_SHORT ) 
									== 0)) {

					URL *l_url = NULL;

					PKI_log_debug( "Trying PRQP RQA -> %s",
								addr );

					if((l_url = URL_new( addr )) == NULL) {
						PKI_log_debug("Memory Error [%s:%d]", __FILE__, __LINE__ );
						continue;
					}

					if( l_url->port <= 0 ) 
						l_url->port = PRQP_DEFAULT_PORT;

					r = PRQP_http_get_resp ( l_url, p, 0);
					if( r == NULL ) {
						PKI_log( PKI_LOG_ERR,
							"Can not get response "
							"from server (%s:%d)!",
								l_url->addr, 
								l_url->port);
						URL_free ( l_url );
					} else {
						/* Exit the cycle */
						PKI_log_debug("Got PRQP response from server");
						URL_free ( l_url );
						break;
					}
				}
                	}
        	}
	}

        fclose(file);

	return( r );
}

