/* TOKEN Object Management Functions */

#include <strings.h>
#include <libpki/pki.h>

#include <sys/types.h>
#include <dirent.h>

/*!
 * \brief Create a new PKI_TOKEN structure.
 *
 * Reserves the memory for a new PKI_TOKEN data structure. The returned
 * memory is already zeroized. No token configuration is performed. If
 * you need to configure the token by using a configuration file, please
 * use the PKI_TOKEN_new() function.
 *
 */

PKI_TOKEN *PKI_TOKEN_new_null( void ) {

	PKI_TOKEN *tk = NULL;

	tk = (PKI_TOKEN *) PKI_Malloc ( sizeof(PKI_TOKEN));
	if( !tk ) return (NULL);

	tk->algor  = PKI_ALGOR_get ( PKI_ALGOR_DEFAULT );

	if((tk->otherCerts = PKI_STACK_X509_CERT_new()) == NULL ) {
		PKI_Free( tk );
		return NULL;
	}

	if((tk->trustedCerts = PKI_STACK_X509_CERT_new()) == NULL ) {
		PKI_Free( tk );
		return NULL;
	}

	if(PKI_TOKEN_set_config_dir ( tk, NULL ) == PKI_ERR ) {
		PKI_log_debug("ERROR, can not set config_dir for TOKEN!\n");
	}

	/* Initialize OpenSSL so that it adds all the needed algor and dgst */
	if( PKI_get_init_status() == PKI_STATUS_NOT_INIT ) PKI_init_all();

	return ( tk );
}


/*!
 * \brief Create a new PKI_TOKEN structure and initialize it.
 *
 * Reserves the memory for a new PKI_TOKEN data structure. The returned
 * memory is already zeroized. If the first passed argument is null, the
 * default configuration directory is used (PREFIX/etc). If the second
 * argument is not NULL, the library will try to load the specified config
 * for the token.
 *
 * It returns the pointer to the memory region or NULL in case of error.
*/

PKI_TOKEN *PKI_TOKEN_new( char * config_dir, char *name ) {

	PKI_TOKEN *tk = NULL;

	if((tk = PKI_TOKEN_new_null()) == NULL ) {
		return (NULL);
	}

	if((PKI_TOKEN_init( tk, config_dir, name )) == PKI_ERR ) {
		PKI_log_debug("ERROR::config loading ERROR!\n");
	}

	return ( tk );
}

/*!
 * \brief Set the configuration directory to be used for the TOKEN operations.
 */

int PKI_TOKEN_set_config_dir ( PKI_TOKEN *tk, char * dir ) {

	if( dir ) {
		tk->config_dir = strdup( dir );
	} else {
		tk->config_dir = strdup( PKI_DEFAULT_CONF_DIR );
	}

	return (PKI_OK);
}

/*!
 * \brief Get the configuration directory used for the TOKEN operations.
 */

char * PKI_TOKEN_get_config_dir ( PKI_TOKEN *tk ) {

	if( !tk || !tk->config_dir ) return ("");

	return( tk->config_dir );
}

/*!
 * \brief Frees a PKI_TOKEN data structure
 *
 * This function Frees a PKI_TOKEN memory region. In case the PKI_TOKEN has
 * already initialized pointers, the pointed data is freed.
 */

int PKI_TOKEN_free( PKI_TOKEN *tk ) {

	if( tk == NULL ) return (PKI_ERR);

	if( tk->req ) {
		PKI_X509_REQ_free( tk->req );
		tk->req = NULL;
	}

	if( tk->cacert ) {
		PKI_X509_CERT_free( tk->cacert );
		tk->cacert = NULL;
	}

	if( tk->keypair ) {
		PKI_KEYPAIR_free( tk->keypair, tk->hsm );
		tk->keypair = NULL;
	}

	if( tk->otherCerts ) {
		PKI_X509_CERT *x = NULL;

		while((x = PKI_STACK_X509_CERT_pop( tk->otherCerts )) 
								!= NULL) {
			PKI_X509_CERT_free( x );
		}
		PKI_STACK_X509_CERT_free( tk->otherCerts );
		tk->otherCerts = NULL;
	}

	if( tk->trustedCerts ) {
		PKI_X509_CERT *x = NULL;

		while((x = PKI_STACK_X509_CERT_pop( tk->trustedCerts )) 
								!= NULL) {
			PKI_X509_CERT_free( x );
		}
		PKI_STACK_X509_CERT_free( tk->trustedCerts );
		tk->trustedCerts = NULL;
	}

	if( tk->hsm ) {
		HSM_free (tk->hsm);
		tk->hsm = NULL;
	}

	if( tk->profiles ) {
		PKI_X509_PROFILE *pr = NULL;
		while( (pr = PKI_STACK_X509_PROFILE_pop ( tk->profiles ))
							!= NULL ) {
			PKI_X509_PROFILE_free ( pr );
		}
	}

	if( tk->cert ) {
		PKI_X509_CERT_free( tk->cert );
		tk->cert = NULL;
	}

	if( tk->config_dir ) {
		PKI_Free (tk->config_dir);
		tk->config_dir = NULL;
	}

	if( tk->config ) {
		PKI_CONFIG_free ( tk->config );
		tk->config = NULL;
	}

	PKI_Free( tk );

	return (PKI_OK);
}

/*!
 * \brief Loads a configuration file from the token.d directory.
 *
 * Loads the configuration to be used by the Token. The second parameter is
 * checked against the name of the configuration (not the filename, but the
 * <pki:name></pki:name> field in the file), and the matching is loaded.
 * Each file which filename ends with .xml is checked in the token.d/ dir.
 * The name comparison is case insensitive.
 *
 * The function returns PKI_OK in case of success, or PKI_ERR in case of an
 * error.
 */

int PKI_TOKEN_load_config ( PKI_TOKEN *tk, char *tk_name ) {

	char buff[BUFF_MAX_SIZE];

	char *hsm_name = NULL;
	char *config_file = NULL;
	char *tmp_s = NULL;

	PKI_CONFIG *hsm_config = NULL;

	/* Check input */
	if( !tk ) {
		return (PKI_ERR);
	}

	if( tk->config_dir ) {
		snprintf(buff, BUFF_MAX_SIZE,"%s/%s", 
				tk->config_dir, PKI_DEFAULT_TOKEN_DIR );
	} else {
		snprintf(buff,BUFF_MAX_SIZE, "%s/%s",
				PKI_DEFAULT_CONF_DIR, PKI_DEFAULT_TOKEN_DIR );
	}

	if((config_file = PKI_CONFIG_find( buff, tk_name )) == NULL ) {
		PKI_log_debug ("Can not find config file (%s/%s)!!!\n",
			buff, tk_name );
		return( PKI_ERR );
	}

	if( tk->config != NULL ) {
		PKI_CONFIG_free ( tk->config );
	}

	if((tk->config = PKI_CONFIG_load( config_file )) == NULL ) {
		PKI_log_debug("Can not load config file (%s)\n", config_file);
		PKI_Free ( config_file );
		return( PKI_ERR );
	}

	if(( tmp_s = PKI_CONFIG_get_value ( tk->config, "/tokenConfig/type" ))
				!= NULL ) {
		if( strncmp_nocase( tmp_s, "software", 8 ) == 0 ) {
			tk->type = HSM_TYPE_SOFTWARE;
		} else if( strncmp_nocase( tmp_s, "engine", 6 ) == 0 ) {
			tk->type = HSM_TYPE_ENGINE;
		} else if( strncmp_nocase( tmp_s, "kmf", 3 ) == 0 ) {
			tk->type = HSM_TYPE_KMF;
		} else if( strncmp_nocase( tmp_s, "pkcs11", 6) == 0 ) {
			tk->type = HSM_TYPE_PKCS11;
		} else {
			tk->type = HSM_TYPE_OTHER;
		}
	} else {
		tk->type = HSM_TYPE_SOFTWARE;
	}

	if( (tk->type != HSM_TYPE_SOFTWARE ) && ((hsm_name = 
			PKI_CONFIG_get_value( tk->config, "/tokenConfig/hsm"))
								!= NULL) ) {
		PKI_log_debug("TK::HSM::NAME => %s\n", hsm_name );

		if( tk->config_dir ) {
			snprintf(buff, BUFF_MAX_SIZE,"%s/%s", 
					tk->config_dir, PKI_DEFAULT_HSM_DIR );
		} else {
			snprintf(buff,BUFF_MAX_SIZE, "%s/%s",
					tk->config_dir, PKI_DEFAULT_HSM_DIR );
		}

		if( (tk->hsm = HSM_new ( buff, hsm_name )) == NULL ) {
			PKI_log_debug("TK::HSM::Can not init %s HSM!\n", 
								hsm_name);
			return(PKI_ERR);
		}
	} else {
		PKI_log_debug("TK::NO HSM NAME FOUND!\n");
	}

	if(( tmp_s = PKI_CONFIG_get_value( tk->config, "/tokenConfig/keypair" ))
								!= NULL ) {
		char *pwd = NULL;
		char *passin = NULL;
		char *tmp_pwd = NULL;

		/* Initialize OpenSSL so that it adds all the needed algor and dgst */
		// if( PKI_get_init_status() == PKI_STATUS_NOT_INIT ) 
		PKI_init_all();

		if(( pwd = PKI_CONFIG_get_value ( tk->config, 
				"/tokenConfig/password")) == NULL ) {
			if((passin = PKI_CONFIG_get_value ( tk->config,
					"/tokenConfig/passin")) != NULL ) {
				if( strncmp_nocase( passin, "env:", 4) == 0) {
					tmp_pwd = getenv( passin+4 );
					if( tmp_pwd ) {
						pwd = strdup( tmp_pwd);
					} else {
						pwd = strdup( "" );
					}
				} else if (strncmp_nocase( passin, "stdin", 5) == 0 ) {
					pwd = NULL;
				} else {
					pwd = strdup( "" );
				}

				PKI_Free ( passin );
			} else {
				pwd = strdup( "" );
			}
		}

		if ( tk->cred == NULL ) tk->cred = PKI_CRED_new();

		tk->cred->password = strdup ( pwd );
		tk->cred->username = NULL;

		if((PKI_TOKEN_load_keypair ( tk, tmp_s )) == PKI_ERR ) {
			PKI_log_debug("TK::KEY::Unable to "
					"load key URL (%s)\n", tmp_s );
		}
		if( pwd ) PKI_Free ( pwd );
	} else {
		PKI_log_debug("TK::KEY::Warning::No Key Provided!\n");
	}

	if(( tmp_s = PKI_CONFIG_get_value ( tk->config, "/tokenConfig/cert"))
			!= NULL ) {
		if((tk->cert = PKI_X509_CERT_get ( tmp_s )) == NULL ) {
			PKI_log_debug("TK::CERT::ERROR::Can't load %s cert!\n",
								tmp_s );
		}
	}

	if((tmp_s = PKI_CONFIG_get_value ( tk->config, 
					"/tokenConfig/otherCerts")) == NULL ) {
		if((tk->otherCerts = PKI_X509_CERT_STACK_get ( tmp_s )) 
								== NULL ) {
			PKI_log_debug( "TK::OTHERCERTS::ERROR::Can't load %s!\n", tmp_s );
		}
	}

	if((tmp_s = PKI_CONFIG_get_value ( tk->config, 
				"/tokenConfig/trustedCerts")) == NULL ) {
		if((tk->trustedCerts = PKI_X509_CERT_STACK_get ( tmp_s )) 
								== NULL ) {
			PKI_log_debug( "TK::TRUSTEDCERTS::ERROR::Can't load %s!\n", tmp_s );
		}
	}

	return (PKI_OK);
}

/*!
 * \brief Initialize Token properties (load OIDs and PROFILES)
 */

int PKI_TOKEN_init( PKI_TOKEN *tk, char *conf_dir, char *tk_name ) {

	char buff[2048];

	PKI_CONFIG *oids = NULL;
	char *hsm_name = NULL;

	char * homedir = NULL;

	if( !tk ) return (PKI_ERR);

	if( tk->config_dir ) {
		PKI_Free ( tk->config_dir );
		tk->config_dir = NULL;
	}

	if( conf_dir ) {
		/* Set the new Config Dir */
		tk->config_dir = strdup ( conf_dir );

	} else {
		if(( homedir = getenv("HOME")) != NULL ) {
			snprintf( buff, sizeof( buff), "%s/.libpki", 
				homedir );
		} else {
			snprintf( buff, sizeof( buff ), "%s",
				PKI_DEFAULT_CONF_DIR );
		}
		tk->config_dir = strdup ( buff );
	}

	/* Load Configuration Files */
	snprintf( buff, sizeof(buff), "%s/%s",
				tk->config_dir, PKI_DEFAULT_CONF_OID_FILE);

	/* Load the external object Identifiers file */
	if((oids = PKI_CONFIG_load( buff )) != NULL ) {
		tk->oids = oids;
	} else {

		PKI_log_debug("WARNING::can not load oids file!\n");
	}

	/* Load Configuration Files */
	snprintf( buff, BUFF_MAX_SIZE, "%s/%s",
				tk->config_dir, PKI_DEFAULT_PROFILE_DIR);

	if( PKI_TOKEN_load_profiles( tk, buff ) == PKI_ERR ) {
		/* We should check why... */
		PKI_log_debug("INFO, can not load profiles (%s)\n", buff);
	}

	/* Now let's try to load the HSM configuration file */
	if( tk_name ) {
		if(( PKI_TOKEN_load_config( tk, tk_name )) == PKI_ERR ) {
			PKI_log_debug( "INFO, can not load config file!\n");
		} else {
			PKI_log_debug( "OK, config file loaded!\n");
		}

		/* Now let's try to load the token configuration */
		/*
		if( PKI_HSM_load_config ( tk, tk_name ) != PKI_OK ) {
			printf("ERROR, can not load the token config!\n");
			return(PKI_ERR);
		}
		*/
	}

	return ( PKI_OK );
}

/*!
 * \brief Returns a pointer to an Object Identifier structure
 *
 * This function returns a pointer to an Object Identifier if the
 * OID is identified among the ones in the crypto library used or
 * among the configured ones in the objectIdentifiers.xml file that
 * is loaded when the PKI_TOKEN_init() function is used.
 */

PKI_OID *PKI_TOKEN_OID_new ( PKI_TOKEN *tk, char *oid_s ) {
	PKI_OID *ret = NULL;

	/* Here every library should provide it's own function
	   for oid generation: PKI_OID_new() and PKI_OID_free() */
	/*
	   ret = PKI_OID_new( oid_s );
        */
	if( !ret ) {
		ret = PKI_CONFIG_OID_search ( tk->oids, oid_s );
	}

	return (ret);
}

/*!
 * \brief Set the scheme to be used for TOKEN operations
 *
 * When generating new keypair, request or certificates by using the
 * PKI_TOKEN facility, a signature scheme has to be chosen. The default
 * one is RSA, but, depending on the capabilities of the system you
 * may be able to use DSA and ECDSA as well. The latter is currently
 * only available when OpenSSL is used as the crypto backend.
 *
 * NOTE: on some systems the Elliptic Curves (EC) capabilities of OpenSSL
 *       are disabled by default, to enable them you might need to link
 *       libPKI with a different version of OpenSSL libraries
 */

/*
int PKI_TOKEN_set_scheme( PKI_TOKEN *tk, int scheme ) {

	if( !tk ) return PKI_ERR;

	tk->scheme = scheme;

	return( PKI_OK );
}
*/

/*!
 * \brief Set the TOKEN's scheme algorithm via its PKI_ALGOR_ID.
 *
 * When generating signatures (e.g., when issuing a new request or certificate)
 * by using the PKI_TOKEN facility, a signature algorithm has to be chosen. The
 * default one is sha1withRSA, but, depending on the capabilities of the system you
 * may be able to use differnt ones as well. This algorithm is used in combination
 * with the signature scheme (e.g., sha1withRSA or md5withDSA), therefore it must
 * be consistent with it.
 *
 * Possible algoritms are: 
 * -PKI_ALGOR_RSA_MD5
 * -PKI_ALGOR_RSA_MD2
 * -PKI_ALGOR_RSA_SHA1
 * -PKI_ALGOR_DSA_SHA1
 * -PKI_ALGOR_RSA_SHA224
 * -PKI_ALGOR_RSA_SHA256
 * -PKI_ALGOR_RSA_SHA512
 * -PKI_ALGOR_RSA_RIPEMD160
 * -PKI_ALGOR_ECDSA_SHA1
 *
 */

int PKI_TOKEN_set_algor( PKI_TOKEN *tk, PKI_ALGOR_ID algId ) {

	if( !tk ) return (PKI_ERR);

	if( (tk->algor = PKI_ALGOR_get( algId )) == PKI_ALGOR_UNKNOWN ) {
		PKI_log_debug("ERROR, TOKEN_set_algor() -> can not set " 
					"algor %d!", algId );
		return ( PKI_ERR );
	};

	return( PKI_OK );
};

/*!
 * \brief Set the TOKEN's scheme algorithm via its name.
 *
 * When generating signatures (e.g., when issuing a new request or certificate)
 * by using the PKI_TOKEN facility, a signature algorithm has to be chosen. The
 * default one is sha1withRSA, but, depending on the capabilities of the system you
 * may be able to use differnt ones as well. This algorithm is used in combination
 * with the signature scheme (e.g., sha1withRSA or md5withDSA), therefore it must
 * be consistent with it.
 *
 * Possible algoritms names are: 
 *   RSA-MD2
 *   RSA-MD4
 *   RSA-MD5
 *   RSA-SHA1
 *   DSA-SHA1
 *   RSA-SHA224
 *   RSA-SHA256
 *   RSA-SHA512
 *   RSA-RIPEMD160
 */

int PKI_TOKEN_set_algor_by_name( PKI_TOKEN *tk, char *alg_name ) {

	if( !tk ) return (PKI_ERR);

	if( (tk->algor = PKI_ALGOR_get_by_name( alg_name )) == NULL ) {
		return ( PKI_ERR );
	}

	return( PKI_OK );
}

/*!
 * \brief Generates a PKI_KEYPAIR and store it in a PKI_TOKEN.
 *
 * This function assumes a new PKI_TOKEN data structure is passed
 * together with the PKI_SCHEME and the number of bit for the new
 * key. It returns PKI_OK in case of success or PKI_ERR if an
 * error occurs.
 *
 * The label is needed by some HSM(s).
 */

int PKI_TOKEN_new_keypair ( PKI_TOKEN *tk, int bits, char *label ) {

	PKI_KEYPAIR *p = NULL;
	PKI_SCHEME_ID scheme = PKI_SCHEME_UNKNOWN;

	if(!tk) return (PKI_ERR);

	if((scheme = PKI_ALGOR_get_scheme ( tk->algor )) 
						== PKI_SCHEME_UNKNOWN ) {
		/* Error, scheme not recognized! */
		PKI_log_debug( "PKI_TOKEN_new_keypair() -> returned scheme "
					"unknown!");
		return ( PKI_ERR );
	}

	if((p = PKI_KEYPAIR_new( scheme, bits, label, 
					tk->cred, tk->hsm )) == NULL ) {
		return(PKI_ERR);
	};

	if( tk->keypair ) PKI_KEYPAIR_free ( tk->keypair, tk->hsm );

	tk->keypair = p;

	return(PKI_OK);
}

/*!
 * \brief Returns the PKI_KEYPAIR of a PKI_TOKEN
 *
 * Use this function to get a reference (not a copy) to the PKI_KEYPAIR
 * of the PKI_TOKEN.
 *
 * The function returns the pointer or NULL in case of error of if no
 * PKI_KEYPAIR has been assigned to the PKI_TOKEN yet. 
 */

PKI_KEYPAIR *PKI_TOKEN_get_keypair ( PKI_TOKEN *tk ) {
	if(!tk) return (NULL);

	return (tk->keypair);
}

/*!
 * \brief Set the PKI_TOKEN certificate
 *
 * Use this function to set the certificate of a PKI_TOKEN. The certificate
 * is automatically freed when the PKI_TOKEN_free() function is used.
 * The function returns PKI_OK in case of success or PKI_ERR otherwise.
 */

int PKI_TOKEN_set_cert ( PKI_TOKEN *tk, PKI_X509_CERT *x ) {

	if( !tk || !x ) return 0;

	if( tk->cert ) {
		PKI_X509_CERT_free( tk->cert );
		tk->cert = NULL;
	}

	tk->cert = x;

	return 1;
}

/*!
 * \brief Returns the certificate of a PKI_TOKEN
 *
 * Use this function to get a reference (not a copy) to the certificate 
 * of the PKI_TOKEN.
 *
 * The function returns the pointer or NULL in case of error of if no
 * certificate has been assigned to the PKI_TOKEN yet. 
 */

PKI_X509_CERT * PKI_TOKEN_get_cert ( PKI_TOKEN *tk ) {
	if( !tk ) return (NULL);

	return tk->cert;
}

/*!
 * \brief Set the PKI_TOKEN CA certificate
 *
 * Use this function to set the CA certificate of a PKI_TOKEN. The certificate
 * is automatically freed when the PKI_TOKEN_free() function is used.
 * The function returns PKI_OK in case of success or PKI_ERR otherwise.
 */

int PKI_TOKEN_set_cacert ( PKI_TOKEN *tk, PKI_X509_CERT *x ) {

	if( !tk || !x ) return PKI_ERR;

	if( tk->cacert ) {
		PKI_X509_CERT_free( tk->cert );
		tk->cert = NULL;
	}

	tk->cacert = x;

	return PKI_OK;
}

/*!
 * \brief Returns the CA certificate of a PKI_TOKEN
 *
 * Use this function to get a reference (not a copy) to the CA certificate 
 * of the PKI_TOKEN.
 *
 * The function returns the pointer or NULL in case of error of if no
 * CA certificate has been assigned to the PKI_TOKEN yet. 
 */

PKI_X509_CERT * PKI_TOKEN_get_cacert ( PKI_TOKEN *tk ) {
	if( !tk ) return (NULL);

	return tk->cacert;
}

/*!
 * \brief Set the PKI_TOKEN stack of additional certificates
 *
 * Use this function to set a stack of certificate to a PKI_TOKEN. The stack
 * is automatically freed when the PKI_TOKEN_free() function is used.
 * The additional certificates are used for validation purposes when a verify
 * of the PKI_TOKEN certificate is performed.
 *
 * The function returns PKI_OK in case of success or PKI_ERR otherwise.
 */

int PKI_TOKEN_set_otherCerts ( PKI_TOKEN *tk, PKI_X509_CERT_STACK *stack ) {
	if(!tk || !stack ) return (PKI_ERR);

	if( tk->otherCerts ) {
		PKI_STACK_X509_CERT_free ( stack );
		tk->otherCerts = NULL;
	}
	tk->otherCerts = stack;

	return(PKI_OK);

}

/*!
 * \brief Set the PKI_TOKEN stack of trusted certificates
 *
 * Use this function to set a stack of trusted certs to a PKI_TOKEN. The stack
 * is automatically freed when the PKI_TOKEN_free() function is used.
 * The additional certificates are used for validation purposes when a verify
 * of the PKI_TOKEN certificate is performed.
 *
 * The function returns PKI_OK in case of success or PKI_ERR otherwise.
 */

int PKI_TOKEN_set_trustedCerts ( PKI_TOKEN *tk, PKI_X509_CERT_STACK *stack ) {
	if(!tk || !stack ) return (PKI_ERR);

	if( tk->trustedCerts ) {
		PKI_STACK_X509_CERT_free ( stack );
		tk->trustedCerts = NULL;
	}
	tk->trustedCerts = stack;

	return(PKI_OK);

}

/*!
 * \brief Returns the stack of certificates (chain) of a PKI_TOKEN
 *
 * Use this function to get a reference (not a copy) to the stack of
 * certificates of the PKI_TOKEN.
 *
 * The function returns the pointer to the PKI_X509_CERT_STACK or NULL
 * in case of error.
 */

PKI_X509_CERT_STACK * PKI_TOKEN_get_otherCerts ( PKI_TOKEN *tk ) {
	if( !tk ) return (NULL);

	return tk->otherCerts;
}

/*!
 * \brief Returns the stack of trusted certificates (chain) of a PKI_TOKEN
 *
 * Use this function to get a reference (not a copy) to the stack of
 * tursted certificates (Trust Anchors) of the PKI_TOKEN.
 *
 * The function returns the pointer to the PKI_X509_CERT_STACK or NULL
 * in case of error.
 */

PKI_X509_CERT_STACK * PKI_TOKEN_get_trustedCerts ( PKI_TOKEN *tk ) {
	if( !tk ) return (NULL);

	return tk->trustedCerts;
}

/*!
 * \brief Set the credentials to be used when retrieving/using the KEYPAIR.
 *
 * Use this function to set the credentials to be used by the PKI_TOKEN. The
 * credentials are automatically freed when the PKI_TOKEN_free() function is used.
 * The function returns PKI_OK in case of success or PKI_ERR otherwise.
 */

int PKI_TOKEN_set_cred ( PKI_TOKEN *tk, PKI_CRED *cred ) {

	if(!tk || !cred) return (PKI_ERR);

	if( tk->cred ) PKI_CRED_free (tk->cred);
	tk->cred = cred;

	return(PKI_OK);
}

/*!
 * \brief Returns the credentials of a PKI_TOKEN
 *
 * Use this function to get a reference (not a copy) to the credentials 
 * of the PKI_TOKEN.
 *
 * The function returns the pointer or NULL in case of error of if no
 * credentials has been assigned to the PKI_TOKEN yet. 
 */

PKI_CRED * PKI_TOKEN_get_cred ( PKI_TOKEN *tk ) {
	if( !tk ) return (NULL);

	return tk->cred;
}

/*!
 * \brief Set the KEYPAIR to be used in the KEYPAIR.
 *
 * Use this function to set the credentials to be used by the PKI_TOKEN. The
 * credentials are automatically freed when the PKI_TOKEN_free() function is used.
 * The function returns PKI_OK in case of success or PKI_ERR otherwise.
 */

int PKI_TOKEN_set_keypair ( PKI_TOKEN *tk, PKI_KEYPAIR *pkey ) {

	if( !tk || !pkey ) return (PKI_ERR);

	if( tk->keypair ) {
		PKI_KEYPAIR_free( tk->keypair, tk->hsm );
		tk->keypair = NULL;
	}

	tk->keypair = pkey;

	return (PKI_OK);
}

/*!
 * \brief Get the CA certificate from a URL and assigns it to the PKI_TOKEN
 *
 * This function loads the CA certificate from a URL and assigns it to the
 * PKI_TOKEN structure. The certificate data structure is automatically
 * freed when the PKI_TOKEN_free() function is used.
 *
 * The function returns PKI_OK in case of success, otherwise it returns
 * PKI_ERR.
 */

int PKI_TOKEN_load_cacert ( PKI_TOKEN *tk, char *url_string ) {

	URL *url = NULL;

	if( !tk || !url_string ) return PKI_ERR;

	if((tk->cacert = PKI_X509_CERT_get( url_string )) == NULL ) {
		/* Can not load the certificate from the given URL */
		URL_free( url );
		return PKI_ERR;
	}

	return(PKI_OK);
}


/*!
 * \brief Get a certificate from a URL and assigns it to the PKI_TOKEN
 *
 * This function loads a certificate from a URL and assigns it to the
 * PKI_TOKEN structure. The certificate data structure is automatically
 * freed when the PKI_TOKEN_free() function is used.
 *
 * The function returns PKI_OK in case of success, otherwise it returns
 * PKI_ERR.
 */

int PKI_TOKEN_load_cert ( PKI_TOKEN *tk, char *url_string ) {

	URL *url = NULL;

	if( !tk || !url_string ) return PKI_ERR;

	if((tk->cert = PKI_X509_CERT_get( url_string )) == NULL ) {
		/* Can not load the certificate from the given URL */
		URL_free( url );
		return PKI_ERR;
	}

	return(PKI_OK);
}

/*!
 * \brief Get a chain of TRUSTED certificates from a URL and assigns them to
 *        the PKI_TOKEN (CA Certificates)
 *
 * This function loads a set of certificates from a URL and assigns it to the
 * PKI_TOKEN structure. The stack of certificates is automatically
 * freed when the PKI_TOKEN_free() function is used.
 *
 * The function returns PKI_OK in case of success, otherwise it returns
 * PKI_ERR.
 */

int PKI_TOKEN_load_trustedCerts( PKI_TOKEN *tk, char *url_string ) {

	PKI_X509_CERT_STACK *cert = NULL;
	URL *url = NULL;
	int i = 0;

	if( !tk || !url_string ) return PKI_ERR;

	if((cert = PKI_X509_CERT_STACK_get( url_string )) == NULL ) {
		/* Can not load the certificate from the given URL */
		URL_free( url );
		return PKI_ERR;
	}

	/* attach the stack to the TOKEN structure */
	if( (i = PKI_STACK_X509_CERT_elements( cert )) > 0 ) {
		tk->trustedCerts = cert;
	}

	/* On success the number of certificates in the chain is
	   returned */
	return(PKI_OK);
}

/*!
 * \brief Get a chain of certificates from a URL and assigns them to
 *        the PKI_TOKEN (Not Trust Anchors - i.e., trusted CA certs)
 *
 * This function loads a set of certificates from a URL and assigns it to the
 * PKI_TOKEN structure. The stack of certificates is automatically
 * freed when the PKI_TOKEN_free() function is used.
 *
 * The function returns PKI_OK in case of success, otherwise it returns
 * PKI_ERR.
 */

int PKI_TOKEN_load_otherCerts( PKI_TOKEN *tk, char *url_string ) {

	PKI_X509_CERT_STACK *cert = NULL;
	URL *url = NULL;
	int i = 0;

	if( !tk || !url_string ) return PKI_ERR;

	if((cert = PKI_X509_CERT_STACK_get( url_string )) == NULL ) {
		/* Can not load the certificate from the given URL */
		URL_free( url );
		return PKI_ERR;
	}

	/* attach the stack to the TOKEN structure */
	if( (i = PKI_STACK_X509_CERT_elements( cert )) > 0 ) {
		tk->otherCerts = cert;
	}

	/* On success the number of certificates in the chain is
	   returned */
	return(PKI_OK);
}

/*!
 * \brief Get a PKI_KEYPAIR from a URL and assigns it to the PKI_TOKEN
 *
 * This function loads a keypair from a URL and assigns it to the
 * PKI_TOKEN structure. The keypair data structure is automatically
 * freed when the PKI_TOKEN_free() function is used.
 *
 * The function returns PKI_OK in case of success, otherwise it returns
 * PKI_ERR.
 */

int PKI_TOKEN_load_keypair ( PKI_TOKEN *tk, char *url_string ) {
	PKI_KEYPAIR *pkey = NULL;
	URL *url = NULL;

	if(!tk | !url_string) return PKI_ERR;

	if(( url = getParsedUrl( url_string)) == NULL ) {
		return PKI_ERR;
	}

	if((pkey = PKI_KEYPAIR_get_url( url, tk->cred, tk->hsm )) == NULL ) {
		/* Can not load the certificate from the given URL */
		URL_free( url );
		return (PKI_ERR);
	}

	tk->keypair = pkey;

	return (PKI_OK);
}

/* Save data to URL pointers */
int PKI_TOKEN_write ( PKI_TOKEN *tk, char *url_string, int format) {

	int ret = PKI_OK;
	URL *url = NULL;

	/* This will write a PKI_TOKEN in an .p12 file */
	if(!tk || !url_string) return (PKI_ERR);

	if((url = URL_new( url_string )) == NULL ) {
		return( PKI_ERR );
	}

	ret = PKI_PKCS12_TOKEN_write( tk, url, format );

	/*
	switch (format) {
		case FORMAT_PKCS12_DER:
			ret = PKI_P12TOKEN_write( tk, url, FORMAT_ASN1 );
			break;
		case PKCS12_PEM:
			ret = PKI_P12TOKEN_write( tk, url, FORMAT_PEM );
			break;
		case XML:
		default:
			ret = PKI_XML_TOKEN_write( tk, url, FORMAT_XML );
	}
	*/

	if( url ) URL_free ( url );
	return ( ret );
}

int PKI_TOKEN_write_cert ( PKI_TOKEN *tk, char *url_string, int format ) {

	if( !tk || !tk->cert || !url_string ) return ( PKI_ERR );

	return PKI_X509_CERT_write( tk->cert, format, url_string );

}

int PKI_TOKEN_write_keypair ( PKI_TOKEN *tk, char *url_string, int format ) {
	return PKI_ERR;
}

int PKI_TOKEN_write_trustedCerts( PKI_TOKEN *tk, char *url_string, int format) {

	URL *url = NULL;

	if( !tk || !tk->cert || !url_string ) return ( PKI_ERR );

	url = URL_new( url_string);
	if( !url ) return (PKI_ERR);

	if( url->proto == URI_PROTO_FILE ) {
		return PKI_X509_CERT_STACK_write_file( tk->trustedCerts, 
							format, url->addr);
	} else{
		PKI_log_debug("Cert write::Protocol not supported %s!\n",
			url_string );
	}
	
	return ( PKI_ERR );
}

int PKI_TOKEN_write_otherCerts ( PKI_TOKEN *tk, char *url_string, int format ) {

	URL *url = NULL;

	if( !tk || !tk->cert || !url_string ) return ( PKI_ERR );

	url = URL_new( url_string);
	if( !url ) return (PKI_ERR);

	if( url->proto == URI_PROTO_FILE ) {
		return PKI_X509_CERT_STACK_write_file( tk->otherCerts, 
							format, url->addr);
	} else{
		PKI_log_debug("Cert write::Protocol not supported %s!\n",
			url_string );
	}
	
	return ( PKI_ERR );
}

int PKI_TOKEN_write_req ( PKI_TOKEN *tk, char *url_string, int format ) {

	URL *url = NULL;

	if( !tk || !url_string ) {
		PKI_log_debug("ERROR, wrong parameters!\n");
		return ( PKI_ERR );
	}

	if( !tk->req ) {
		PKI_log_debug("ERROR, no req to save!\n");
		return ( PKI_ERR );
	}

	url = URL_new( url_string);
	if( !url ) {
		PKI_log_debug("ERROR, can not parse URL %s\n", url_string);
		return (PKI_ERR);
	}

	if( url->proto == URI_PROTO_FILE ) {
		return PKI_X509_REQ_write_url( tk->req, format, url );
	} else{
		PKI_log_debug("REQ WRITE::Protocol not supported %s!\n",
			url_string );
	}
	
	return ( PKI_ERR );
}


/* TOKEN operations */
int PKI_TOKEN_new_req(PKI_TOKEN *tk, char *subject, char *profile_s ) {

	PKI_X509_PROFILE *req_profile = NULL;
	PKI_DIGEST_ALG *digest = NULL;
	int i = 0;

	if( !tk || !tk->keypair ) {
		PKI_log_debug("ERROR, no token or keypair passed to PKI_TOKEN_new_req()!\n");
		return (PKI_ERR);
	}

	if( tk->req ) {
		PKI_log_debug("ERROR, a Request already exists in TOKEN when calling PKI_TOKEN_new_req()!\n");
		return (PKI_ERR);
	}

	if( profile_s ) {
		PKI_X509_PROFILE *tmp_profile = NULL;
		char *prof_name = NULL;

		if( !tk->profiles ) {
			PKI_log_debug("PKI_TOKEN_new_req() -> no profiles!\n");
			return (PKI_ERR);
		}

		for( i=0; i < PKI_STACK_X509_PROFILE_elements(tk->profiles);i++) {
			tmp_profile = PKI_STACK_X509_PROFILE_get_num ( 
							tk->profiles, i );
			prof_name = PKI_X509_PROFILE_get_name( tmp_profile );
			if( !prof_name ) {
				PKI_log_debug("WARNING, No profile name for "
					"profile number %d\n",i);
				continue;
			}

			if( strcmp_nocase( profile_s, prof_name ) == 0 ) {
				req_profile = tmp_profile;
				break;
			}
		}

		if( !req_profile ) {
			/* Error, the requested profile does not
			   exists! */
			PKI_log_debug("ERROR, no matching profile found!\n");
			return (PKI_ERR);
		}
	}
	PKI_log_debug("INFO, Creating new req (%s)\n", subject);

	if((digest = PKI_ALGOR_get_digest ( tk->algor )) 
					== PKI_DIGEST_ALG_UNKNOWN ) {
		/* Error, digest algorithm not supported or recognized! */
		return ( PKI_ERR );
	}

	tk->req = PKI_X509_REQ_new( tk->keypair, subject, req_profile,
			tk->oids, digest, tk->hsm ); 

	if( !tk->req ) return (PKI_ERR);

	return ( PKI_OK );
}

int PKI_TOKEN_set_req( PKI_TOKEN *tk, PKI_X509_REQ *req ) {
	if (!tk || !req ) return (PKI_ERR);

	tk->req = req;

	return (PKI_OK);
}

PKI_X509_PROFILE *PKI_TOKEN_search_profile( PKI_TOKEN *tk, char *profile_s ) {

	PKI_X509_PROFILE *tmp_profile = NULL;
	PKI_X509_PROFILE *ret = NULL;
	char *prof_name = NULL;
	int i = 0;

	if( !tk || !tk->profiles || !profile_s ) return (NULL);

	for( i=0; i < PKI_STACK_X509_PROFILE_elements( tk->profiles ); i++) {
		tmp_profile = PKI_STACK_X509_PROFILE_get_num(tk->profiles, i);

		if((prof_name = PKI_X509_PROFILE_get_name( tmp_profile ))
								== NULL) {
			continue;
		}

		if( strcmp_nocase( profile_s, prof_name ) == 0 ) {
			ret = tmp_profile;
			break;
		}
	}

	return (ret);

}

int PKI_TOKEN_self_sign (PKI_TOKEN *tk, char *subject, char *serial,
							char *profile_s ) {

	PKI_X509_PROFILE *cert_profile = NULL;

	if( !tk || !tk->keypair ) return (PKI_ERR);

	if( tk->cert ) {
		/* ERROR, a Token Certificate already exists! */
		PKI_log_debug("ERROR, A cert already exists in token when calling PKI_TOKEN_self_sign()!\n");
		return (PKI_ERR);
	}

	if( profile_s ) {
		if((cert_profile = PKI_TOKEN_search_profile (tk, profile_s ))
							== NULL) {;

			/* Error, the requested profile does not
			   exists! */
			PKI_log_debug("ERROR, Requested profile (%s) not found when self-signing cert!\n", profile_s);
			return (PKI_ERR);
		}
	}

	tk->cert = PKI_X509_CERT_new ( NULL, tk->keypair, tk->req, subject,
			serial, cert_profile, tk->algor, tk->oids, tk->hsm );

	if( !tk->cert ) return (PKI_ERR);

	return (PKI_OK);
}

PKI_X509_CERT * PKI_TOKEN_issue_cert(PKI_TOKEN *tk, char *subject, char *serial,
					PKI_X509_REQ *req, char *profile_s ) {

	PKI_X509_PROFILE *cert_profile = NULL;

	if( !tk || !tk->keypair ) return (PKI_ERR);

	if( tk->cert ) {
		/* ERROR, a Token Certificate already exists! */
		return (PKI_ERR);
	}

	if( profile_s ) {
		if((cert_profile = PKI_TOKEN_search_profile (tk, profile_s ))
							== NULL) {;

			/* Error, the requested profile does not
			   exists! */
			return (PKI_ERR);
		}
	}

	return PKI_X509_CERT_new ( NULL, tk->keypair, req, subject,
			serial, cert_profile, tk->algor, tk->oids, tk->hsm );

}

int PKI_TOKEN_load_profiles ( PKI_TOKEN *tk, char *urlStr ) {

	// struct dirent **namelist = NULL;
	// struct direct dirbuf;
        struct dirent *dd = NULL;
	DIR *dirp = NULL;
	URL *url = NULL;
	char * fullpath = NULL;

	/* Check input */
	if( !tk || !urlStr ) {
		PKI_log_debug("ERROR, one of tk (%p) and url (%p) are empty"
			"when calling PKI_TOKEN_load_profiles()!\n",
				tk, urlStr);
		return (PKI_ERR);
	}

	if(( url = URL_new ( urlStr )) == NULL ) {
		PKI_log_debug("ERROR, can not parse uri %s in when loading "
					"profiles!\n", urlStr);
		return (PKI_ERR);
	}

	if((tk->profiles = PKI_STACK_X509_PROFILE_new()) == NULL ) {
		PKI_log_debug("ERROR, can not create a new PROFILE STACK!\n");
		return (PKI_ERR);
	}

	if( url->proto != URI_PROTO_FILE ) {
		PKI_log_debug("ERROR, only file:// is currently supported for"
		 			"profiles loading!\n");
		return (PKI_ERR);
	}

	if((dirp = opendir( url->addr )) == NULL ) {
		PKI_log_debug("ERROR, can not load directory %s!\n", 
								url->addr );
		return (PKI_ERR);
	} else {
		while(( dd = readdir( dirp )) != NULL ) {
			long len;
			char *filename = NULL;

			filename = dd->d_name;
			len = (long) strlen( filename );

			if( strcmp( ".xml", filename +len -4 )) {
				continue;
			} else {
				long fullsize = 0;

				fullsize = strlen(url->addr) + 
							strlen( filename ) + 2;

				if((fullsize = strlen(url->addr) + 
					strlen( filename ) + 2) > BUFF_MAX_SIZE) {
					PKI_log_debug("ERROR, filename too long!\n");
					continue;
				}
				
				if((fullpath = PKI_Malloc(fullsize)) == NULL) {
					PKI_log_debug("ERROR, name too long!\n");
					continue;
				};

				snprintf(fullpath, fullsize,
						"%s/%s", url->addr, filename );

				PKI_STACK_X509_PROFILE_push( tk->profiles,
					PKI_X509_PROFILE_load ( fullpath ));

				PKI_Free ( fullpath );
				fullpath = NULL;
			}
		}
		closedir( dirp );
	}
	if( url ) URL_free (url);

	return (PKI_OK);
}

/*
int PKI_TOKEN_set_cert_profile( PKI_TOKEN *tk, PKI_X509_PROFILE *cert_prof ) {
	return PKI_ERR;
}
int PKI_TOKEN_set_req_profile( PKI_TOKEN *tk, PKI_X509_PROFILE *req_prof ) {
	return PKI_ERR;
}
*/


